# Copyright 2020. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import os
import csv
import h5py
import numpy as np

from .core import SortOrder, csv_headers, col_population, find_conversion
from .core import MPI_rank, comm_barrier
from bmtk.utils.sonata.utils import add_hdf5_magic, add_hdf5_version


def write_sonata(path, spiketrain_reader, mode='w', sort_order=SortOrder.none, units='ms',
                 population_renames=None, **kwargs):
    path_dir = os.path.dirname(path)
    if MPI_rank == 0 and path_dir and not os.path.exists(path_dir):
        os.makedirs(path_dir)

    spiketrain_reader.flush()
    comm_barrier()

    populations = spiketrain_reader.populations
    spikes_root = None
    if MPI_rank == 0:
        h5 = h5py.File(path, mode=mode)
        add_hdf5_magic(h5)
        add_hdf5_version(h5)
        spikes_root = h5.create_group('/spikes') if '/spikes' not in h5 else h5['/spikes']

    for pop_name in populations: # metrics.keys():
        if MPI_rank == 0 and pop_name in spikes_root:
            # Problem if file already contains /spikes/<pop_name> # TODO: append new data to old spikes?!?
            raise ValueError('sonata file {} already contains a spikes group {}, '.format(path, pop_name) +
                             'skiping(use option mode="w" to overwrite)')

        pop_df = spiketrain_reader.to_dataframe(populations=pop_name, with_population_col=False, sort_order=sort_order,
                                                on_rank='root')
        if MPI_rank == 0:
            spikes_pop_grp = spikes_root.create_group(pop_name)
            if sort_order != SortOrder.unknown:
                spikes_pop_grp.attrs['sorting'] = sort_order.value

            spikes_pop_grp.create_dataset('timestamps', data=pop_df['timestamps'])
            spikes_pop_grp['timestamps'].attrs['units'] = spiketrain_reader.units()
            spikes_pop_grp.create_dataset('node_ids', data=pop_df['node_ids'])
    comm_barrier()


def write_sonata_itr(path, spiketrain_reader, mode='w', sort_order=SortOrder.none, units='ms', population_renames=None,
                     **kwargs):
    path_dir = os.path.dirname(path)
    if MPI_rank == 0 and path_dir and not os.path.exists(path_dir):
        os.makedirs(path_dir)

    spiketrain_reader.flush()
    comm_barrier()

    conv_factor = find_conversion(spiketrain_reader.units, units)
    if MPI_rank == 0:
        h5 = h5py.File(path, mode=mode)
        add_hdf5_magic(h5)
        add_hdf5_version(h5)
        spikes_root = h5.create_group('/spikes') if '/spikes' not in h5 else h5['/spikes']

    population_renames = population_renames or {}
    for pop_name in spiketrain_reader.populations:
        n_spikes = spiketrain_reader.n_spikes(pop_name)
        if n_spikes <= 0:
            continue

        if MPI_rank == 0:
            spikes_grp = spikes_root.create_group('{}'.format(population_renames.get(pop_name, pop_name)))
            if sort_order != SortOrder.unknown:
                spikes_grp.attrs['sorting'] = sort_order.value

            timestamps_ds = spikes_grp.create_dataset('timestamps', shape=(n_spikes,), dtype=np.float64)
            timestamps_ds.attrs['units'] = units
            node_ids_ds = spikes_grp.create_dataset('node_ids', shape=(n_spikes,), dtype=np.uint64)

        for i, spk in enumerate(spiketrain_reader.spikes(populations=pop_name, sort_order=sort_order)):
            if MPI_rank == 0:
                timestamps_ds[i] = spk[0]*conv_factor
                node_ids_ds[i] = spk[2]

    comm_barrier()


def write_csv(path, spiketrain_reader, mode='w', sort_order=SortOrder.none, include_header=True,
              include_population=True, units='ms', **kwargs):
    path_dir = os.path.dirname(path)
    if MPI_rank == 0 and path_dir and not os.path.exists(path_dir):
        os.makedirs(path_dir)

    df = spiketrain_reader.to_dataframe(sort_order=sort_order, on_rank='root')

    if MPI_rank == 0:
        df[['timestamps', 'population', 'node_ids']].to_csv(path, header=include_header, index=False, sep=' ')

    comm_barrier()


def write_csv_itr(path, spiketrain_reader, mode='w', sort_order=SortOrder.none, include_header=True,
                  include_population=True, units='ms', **kwargs):
    path_dir = os.path.dirname(path)
    if MPI_rank == 0 and path_dir and not os.path.exists(path_dir):
        os.makedirs(path_dir)

    conv_factor = find_conversion(spiketrain_reader.units, units)
    cols_to_print = csv_headers if include_population else [c for c in csv_headers if c != col_population]
    if MPI_rank == 0:
        f = open(path, mode=mode)
        csv_writer = csv.writer(f, delimiter=' ')
        if include_header:
            csv_writer.writerow(cols_to_print)

    for spk in spiketrain_reader.spikes(sort_order=sort_order):
        if MPI_rank == 0:
            ts = spk[0]*conv_factor
            c_data = [ts, spk[1], spk[2]] if include_population else [ts, spk[2]]
            csv_writer.writerow(c_data)

    comm_barrier()



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import os
import h5py
import numpy as np

from bmtk.utils.sonata.utils import add_hdf5_magic, add_hdf5_version

class CurrentWriterv01(object):
    """Used to save current currents to the described hdf5 format.
    """

    class DataTable(object):
        """A small struct to keep track of different data (and buffer) tables"""
        def __init__(self):
            # If buffering data, buffer_block will be an in-memory array and will write to data_block during when
            # filled. If not buffering buffer_block is an hdf5 dataset and data_block is ignored
            self.data_block = None
            self.buffer_block = None

    def __init__(self, file_name, num_currents, units=None, tstart=0.0, tstop=1.0, dt=0.01, n_steps=None,
                 buffer_size=0, buffer_data=True, **kwargs):
        self._units = units
        self._tstart = tstart
        self._tstop = tstop
        self._dt = dt

        self._num_currents = num_currents

        self._n_steps = n_steps

        self._buffer_size = buffer_size
        self._buffer_data = buffer_data and buffer_size>0
        self._data_block = self.DataTable()
        self._last_save_indx = 0  # for buffering, used to keep track of last timestep data was saved to disk

        self._file_name = file_name
        self._create_h5file()

        self._buffer_block_size = 0
        self._total_steps = 0

        self._is_initialized = False

    def set_units(self, val):
        self._units = val

    def units(self):
        return self._units

    def set_tstart(self, val):
        self._tstart = val

    def tstart(self):
        return self._tstart

    def set_tstop(self, val):
        self._tstop = val

    def tstop(self):
        return self._tstop

    def set_dt(self, val):
        self._dt = val

    def dt(self):
        return self._dt

    def n_steps(self):
        if self._n_steps is None:
            self._n_steps = int((self._tstop - self._tstart) / self._dt)
        return self._n_steps

    def _create_h5file(self):
        fdir = os.path.dirname(os.path.abspath(self._file_name))
        if not os.path.exists(fdir):
            os.mkdir(fdir)
        self.h5_base = h5py.File(self._file_name, 'w')
        add_hdf5_version(self.h5_base)
        add_hdf5_magic(self.h5_base)

    def initialize(self, **kwargs):
        if self._is_initialized:
            return

        n_steps = self.n_steps()
        if n_steps <= 0:
            raise Exception('A non-zero positive integer num-of-steps is required to initialize the current report.'
                            'Please specify report length using the n_steps parameters (or using appropiate tstop,'
                            'tstart, and dt).')

        #self._calc_offset()
        base_grp = self.h5_base

        self._total_steps = n_steps
        self._buffer_block_size = self._buffer_size

        if self._buffer_data:
            # Set up in-memory block to buffer recorded variables before writing to the dataset.
            self._data_block.buffer_block = np.zeros((self._buffer_size, self._num_currents), dtype=np.float)

            self._data_block.data_block = base_grp.create_dataset('data', shape=(self.n_steps(), self._num_currents),
                                                                  dtype=np.float, chunks=True)

            if self._units is not None:
                self._data_block.data_block.attrs['units'] = self._units

        else:
            # Since we are not buffering data, we just write directly to the on-disk dataset.
            self._data_block.buffer_block = base_grp.create_dataset('data', shape=(self.n_steps(), self._n_segments_all),
                                                               dtype=np.float, chunks=True)

            if self._units is not None:
                self._data_block.buffer_block.attrs['units'] = self._units

        self._is_initialized = True

    def record_clamps(self, vals, tstep):
        """Record clamp currents.

        :param vals: list of currents for each clamp
        :param tstep: time step
        """
        self.initialize()
        buffer_block = self._data_block.buffer_block
        update_index = (tstep - self._last_save_indx)
        buffer_block[update_index, :] = vals

    def flush(self):
        """Move data from memory to dataset"""
        if self._buffer_data:
            blk_beg = self._last_save_indx
            blk_end = blk_beg + self._buffer_block_size
            if blk_end > self._total_steps:
                # Need to handle the case that simulation doesn't end on a block step
                blk_end = blk_beg + self._total_steps - blk_beg

            block_size = blk_end - blk_beg
            self._last_save_indx += block_size

            self._data_block.data_block[blk_beg:blk_end, :] = self._data_block.buffer_block[:block_size, :]

    def close(self):
        self.h5_base.close()

    def merge(self):
        pass



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


def convert_rates(rates_file):
    rates_df = pd.read_csv(rates_file, sep=' ', names=['gid', 'time', 'rate'])
    rates_sorted_df = rates_df.sort_values(['gid', 'time'])
    rates_dict = {}
    for gid, rates in rates_sorted_df.groupby('gid'):
        start = rates['time'].iloc[0]
        end = rates['time'].iloc[-1]
        dt = float(end - start)/len(rates)
        rates_dict[gid] = {'start': start, 'end': end, 'dt': dt, 'rates': np.array(rates['rate'])}

    return rates_dict


def firing_rates_equal(rates_file1, rates_file2, err=0.0001):
    trial_1 = convert_rates(rates_file1)
    trial_2 = convert_rates(rates_file2)
    if set(trial_1.keys()) != set(trial_2.keys()):
        return False

    for gid, rates_data1 in trial_1.items():
        rates_data2 = trial_2[gid]
        if rates_data1['dt'] != rates_data2['dt'] or rates_data1['start'] != rates_data2['start'] or rates_data1['end'] != rates_data2['end']:
            return False

        for r1, r2 in zip(rates_data1['rates'], rates_data2['rates']):
            if abs(r1 - r2) > err:
                return False

    return True


def plot_rates_popnet(cell_models_file, rates_file, model_keys=None, save_as=None, show_plot=True):
    """Initial method for plotting popnet output

    :param cell_models_file:
    :param rates_file:
    :param model_keys:
    :param save_as:
    :param show_plot:
    :return:
    """

    pops_df = pd.read_csv(cell_models_file, sep=' ')
    lookup_col = model_keys if model_keys is not None else 'node_type_id'
    pop_keys = {str(r['node_type_id']): r[lookup_col] for _, r in pops_df.iterrows()}

    # organize the rates file by population
    # rates = {pop_name: ([], []) for pop_name in pop_keys.keys()}
    rates_df = pd.read_csv(rates_file, sep=' ', names=['id', 'times', 'rates'])
    for grp_key, grp_df in rates_df.groupby('id'):
        grp_label = pop_keys[str(grp_key)]
        plt.plot(grp_df['times'], grp_df['rates'], label=grp_label)

    plt.legend(fontsize='x-small')
    plt.xlabel('time (s)')
    plt.ylabel('firing rates (Hz)')

    if save_as is not None:
        plt.savefig(save_as)

    if show_plot:
        plt.show()



import numpy as np
from collections import namedtuple

PopulationID = namedtuple('PopulationID', 'node_id population')


class GidPool(object):
    def __init__(self):
        # map from pool-id --> gid
        self._accumulated_offset = 0
        self._pool_offsets = {}

        # map from gid --> pop, node_id
        self._offsets = np.array([0], dtype=np.uint64)
        self._offset2pool_map = {}

    def add_pool(self, name, n_nodes):
        offset_index = len(self._offsets)
        self._offset2pool_map[offset_index] = name
        self._offsets = np.append(self._offsets, np.array([self._accumulated_offset + n_nodes], dtype=np.uint64))

        self._pool_offsets[name] = self._accumulated_offset
        self._accumulated_offset += n_nodes

    def get_gid(self, name, node_id):
        return int(self._pool_offsets[name] + node_id)

    def get_pool_id(self, gid):
        offset_indx = np.searchsorted(self._offsets, gid, 'right')
        node_id = gid - self._offsets[offset_indx-1]
        pool_name = self._offset2pool_map[offset_indx]
        return PopulationID(int(node_id), pool_name)



from bmtk.simulator.pointnet.pyfunction_cache import add_synapse_model


def static_synapse(edge, source_node=None, target_node=None):
    model_params = {
        'model': 'static_synapse',
        'delay': edge.delay,
        'weight': edge.syn_weight(source_node, target_node)
    }

    model_params.update(edge.dynamics_params)
    return model_params


add_synapse_model(static_synapse, 'default', overwrite=False)
add_synapse_model(static_synapse, overwrite=False)


import numpy as np

from .kernel import Kernel3D


class SpatioTemporalFilter(object):
    def __init__(self, spatial_filter, temporal_filter, amplitude=1.):
        self.spatial_filter = spatial_filter
        self.temporal_filter = temporal_filter
        self.amplitude = amplitude

    def get_spatiotemporal_kernel(self, row_range, col_range, t_range=None, threshold=0, reverse=False):
        # TODO: Rename to get_kernel() to match with spatialfilter and temporalfilter
        spatial_kernel = self.spatial_filter.get_kernel(row_range, col_range, threshold=0)
        temporal_kernel = self.temporal_filter.get_kernel(t_range=t_range, threshold=0, reverse=reverse)

        t_range = temporal_kernel.t_range
                
        spatiotemporal_kernel = np.ones((len(temporal_kernel), len(spatial_kernel)))
        spatiotemporal_kernel *= spatial_kernel.kernel[None, :]

        spatiotemporal_kernel *= temporal_kernel.kernel[:, None]
        spatiotemporal_kernel = spatiotemporal_kernel.reshape((np.prod(spatiotemporal_kernel.shape)))
        
        spatial_coord_array = np.empty((len(spatial_kernel), 2))
        spatial_coord_array[:, 0] = spatial_kernel.col_inds
        spatial_coord_array[:, 1] = spatial_kernel.row_inds

        spatiiotemporal_coord_array = np.zeros((len(spatial_kernel)*len(temporal_kernel), 3))
        spatiiotemporal_coord_array[:, 0:2] = np.kron(np.ones((len(temporal_kernel), 1)), spatial_coord_array)
        spatiiotemporal_coord_array[:, 2] = np.kron(temporal_kernel.t_inds, np.ones(len(spatial_kernel)))
        
        col_inds, row_inds, t_inds = map(lambda x: x.astype(np.int64), spatiiotemporal_coord_array.T)
        kernel = Kernel3D(spatial_kernel.row_range, spatial_kernel.col_range, t_range, row_inds, col_inds, t_inds,
                          spatiotemporal_kernel)
        kernel.apply_threshold(threshold)

        kernel.kernel *= self.amplitude
        return kernel
    
    def t_slice(self, t, *args, **kwargs):
        k = self.get_spatiotemporal_kernel(*args, **kwargs)
        return k.t_slice(t)
    
    def show_temporal_filter(self, *args, **kwargs):
        self.temporal_filter.imshow(*args, **kwargs)
            
    def show_spatial_filter(self, *args, **kwargs):
        self.spatial_filter.imshow(*args, **kwargs)
    
    def to_dict(self):
        return {'class': (__name__, self.__class__.__name__),
                'spatial_filter': self.spatial_filter.to_dict(),
                'temporal_filter': self.temporal_filter.to_dict(),
                'amplitude': self.amplitude}



class SimEdge(object):
    @property
    def node_id(self):
        raise NotImplementedError()

    @property
    def gid(self):
        raise NotImplementedError()


class EdgePopulation(object):
    @property
    def source_nodes(self):
        raise NotImplementedError()

    @property
    def target_nodes(self):
        raise NotImplementedError()

    def initialize(self, network):
        raise NotImplementedError()


import numpy as np
import matplotlib.pyplot as plt

def line_plot(evaluate_result, ax=None, show=True, save_file_name=None, xlabel=None, plotstyle=None):

        if ax is None:
            _, ax = plt.subplots(1,1)

        if not plotstyle is None:
            for ((t_range, y_vals), curr_plotstyle) in zip(evaluate_result, plotstyle):   
                ax.plot(t_range, y_vals, curr_plotstyle)
        else:
            for t_range, y_vals in evaluate_result:   
                ax.plot(t_range, y_vals)

        if xlabel is None:
            ax.set_xlabel('Time (Seconds)')
        else:
            ax.set_xlabel(xlabel)
            
        if xlabel is None:
            ax.set_xlabel('Firing Rate (Hz)')
        else:
            ax.set_xlabel(xlabel)
         
        if not save_file_name is None:
            plt.savefig(save_file_name, transparent=True)
            

            
             
        if show == True:
            plt.show() 

def heat_plot(evaluate_result, ax=None, show=True, save_file_name=None, colorbar=True, **kwargs):
    
        if ax is None:
            _, ax = plt.subplots(1,1)

        data = np.empty((len(evaluate_result), len(evaluate_result[0][0])))
        for ii, (t_vals, y_vals) in enumerate(evaluate_result):   
            data[ii,:] = y_vals
            
        cax = ax.pcolor(t_vals, np.arange(len(evaluate_result)), data, **kwargs)
        ax.set_ylim([0,len(evaluate_result)-1])
        ax.set_xlim([t_vals[0], t_vals[-1]])
        ax.set_ylabel('Neuron id')
        ax.set_xlabel('Time (Seconds)')
        
        if colorbar == True:
            plt.colorbar(cax)
         
        if not save_file_name is None:
            plt.savefig(save_file_name, transparent=True)
             
        if show == True:
            plt.show() 
        
        
            

class LGNModel(object):
    def __init__(self, cell_list):
        self.cell_list = cell_list
        
    def evaluate(self, movie, **kwargs):
        return [cell.evaluate(movie, **kwargs) for cell in self.cell_list]

    def __len__(self):
        return len(self.cell_list)



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import os
import logging
from six import string_types

from dipde.internals.internalpopulation import InternalPopulation
from dipde.internals.externalpopulation import ExternalPopulation
from dipde.internals.connection import Connection
import dipde

from bmtk.simulator.core.simulator import Simulator
from . import config as cfg
from . import utils as poputils
import bmtk.simulator.utils.simulation_inputs as inputs
from bmtk.utils.reports.spike_trains import SpikeTrains
from bmtk.utils.io import firing_rates


class PopSimulator(Simulator):
    def __init__(self, graph, dt=0.0001, tstop=0.0, overwrite=True):
        self._graph = graph

        self._tstop = tstop
        self._dt = dt
        self._rates_file = None  # name of file where the output is saved

        self.__population_list = []  # list of all populations, internal and external
        self.__connection_list = []  # list of all connections
        self._dipde_network = None  # reference to dipde.Network object

        self.io = self._graph.io

    @property
    def tstop(self):
        return self._tstop

    @tstop.setter
    def tstop(self, value):
        self._tstop = value

    @property
    def dt(self):
        return self._dt

    @dt.setter
    def dt(self, value):
        self._dt = value

    @property
    def rates_file(self):
        return self._rates_file

    @rates_file.setter
    def rates_file(self, value):
        self._rates_file = value

    @property
    def populations(self):
        return self.__population_list

    @property
    def connections(self):
        return self.__connection_list

    def add_rates_nwb(self, network, nwb_file, trial, force=False):
        """Creates external population firing rates from an NWB file.

        Will iterate through a processing trial of an NWB file by assigning gids the population it belongs too and
        taking the average firing rate.

        This should be done before calling build_cells(). If a population has already been assigned a firing rate an
        error will occur unless force=True.

        :param network: Name of network with external populations.
        :param nwb_file: NWB file with spike rates.
        :param trial: trial id in NWB file
        :param force: will overwrite existing firing rates
        """
        existing_rates = self._rates[network]  # TODO: validate network exists
        # Get all unset, external populations in a network.
        network_pops = self._graph.get_populations(network)
        selected_pops = []
        for pop in network_pops:
            if pop.is_internal:
                continue
            elif not force and pop.pop_id in existing_rates:
                self.io.log_info('Firing rate for {}/{} has already been set, skipping.'.format(network, pop.pop_id))
            else:
                selected_pops.append(pop)

        if selected_pops:
            # assign firing rates from NWB file
            # TODO:
            rates_dict = poputils.get_firing_rate_from_nwb(selected_pops, nwb_file, trial)
            self._rates[network].update(rates_dict)

    def add_rate_hz(self, network, pop_id, rate, force=False):
        """Set the firing rate of an external population.

        This should be done before calling build_cells(). If a population has already been assigned a firing rate an
        error will occur unless force=True.

        :param network: name of network with wanted exteranl population
        :param pop_id: name/id of external population
        :param rate: firing rate in Hz.
        :param force: will overwrite existing firing rates
        """
        self.__add_rates_validator(network, pop_id, force)
        self._rates[network][pop_id] = rate

    def __add_rates_validator(self, network, pop_id, force):
        if network not in self._graph.networks:
            raise Exception('No network {} found in PopGraph.'.format(network))

        pop = self._graph.get_population(network, pop_id)
        if pop is None:
            raise Exception('No population with id {} found in {}.'.format(pop_id, network))
        if pop.is_internal:
            raise Exception('Population {} in {} is not an external population.'.format(pop_id, network))
        if not force and pop_id in self._rates[network]:
            raise Exception('The firing rate for {}/{} already set and force=False.'.format(network, pop_id))

    def _get_rate(self, network, pop):
        """Gets the firing rate for a given population"""
        return self._rates[network][pop.pop_id]

    def build_populations(self):
        """Build dipde Population objects from graph nodes.

        To calculate external populations firing rates, it first see if a population's firing rate has been manually
        set in the graph. Otherwise it attempts to calulate the firing rate from the call to add_rate_hz, add_rates_NWB,
        etc. (which should be called first).
        """
        for network in self._graph.networks:
            for pop in self._graph.get_populations(network):
                if pop.is_internal:
                    dipde_pop = self.__create_internal_pop(pop)

                else:
                    dipde_pop = self.__create_external_pop(pop, self._get_rate(network, pop))

                self.__population_list.append(dipde_pop)
                self.__population_table[network][pop.pop_id] = dipde_pop

    def set_logging(self, log_file):
        # TODO: move this out of the function, put in io class
        if os.path.exists(log_file):
            os.remove(log_file)

        # get root logger
        logger = logging.getLogger()
        for h in list(logger.handlers):
            # remove existing handlers that will write to console.
            logger.removeHandler(h)

        # creates handler that write to log_file
        logging.basicConfig(filename=log_file, filemode='w', level=logging.DEBUG)

    def set_external_connections(self, network_name):
        """Sets the external connections for populations in a given network.

        :param network_name: name of external network with External Populations to connect to internal pops.
        """
        for edge in self._graph.get_edges(network_name):
            # Get source and target populations
            src = edge.source
            source_pop = self.__population_table[src.network][src.pop_id]
            trg = edge.target
            target_pop = self.__population_table[trg.network][trg.pop_id]

            # build a connection.
            self.__connection_list.append(self.__create_connection(source_pop, target_pop, edge))

    def set_recurrent_connections(self):
        """Initialize internal connections."""
        for network in self._graph.internal_networks():
            for edge in self._graph.get_edges(network):
                src = edge.source
                source_pop = self.__population_table[src.network][src.pop_id]
                trg = edge.target
                target_pop = self.__population_table[trg.network][trg.pop_id]
                self.__connection_list.append(self.__create_connection(source_pop, target_pop, edge))

    def run(self, tstop=None):
        # TODO: Check if cells/connections need to be rebuilt.

        # Create the network
        dipde_pops = [p.dipde_obj for p in self._graph.populations]
        dipde_conns = [c.dipde_obj for c in self._graph.connections]
        self._dipde_network = dipde.Network(population_list=dipde_pops, connection_list=dipde_conns)

        if tstop is None:
            tstop = self.tstop

        self.io.log_info("Running simulation.")
        self._dipde_network.run(t0=0.0, tf=tstop, dt=self.dt)
        # TODO: make record_rates optional?
        self.__record_rates()
        self.io.log_info("Finished simulation.")

    def __create_internal_pop(self, params):
        # TODO: use getter methods directly in case arguments are not stored in dynamics params
        # pop = InternalPopulation(**params.dynamics_params)
        pop = InternalPopulation(**params.model_params)
        return pop

    def __create_external_pop(self, params, rates):
        pop = ExternalPopulation(rates, record=False)
        return pop

    def __create_connection(self, source, target, params):
        return Connection(source, target, nsyn=params.nsyns, delays=params.delay, weights=params.weight)

    def __record_rates(self):
        with open(self._rates_file, 'w') as f:
            for pop in self._graph.internal_populations:
                if pop.record:
                    for time, rate in zip(pop.dipde_obj.t_record, pop.dipde_obj.firing_rate_record):
                        f.write('{} {} {}\n'.format(pop.pop_id, time, rate))

    @classmethod
    def from_config(cls, configure, graph):
        # load the json file or object
        if isinstance(configure, string_types):
            config = cfg.from_json(configure, validate=True)
        elif isinstance(configure, dict):
            config = configure
        else:
            raise Exception('Could not convert {} (type "{}") to json.'.format(configure, type(configure)))

        if 'run' not in config:
            raise Exception('Json file is missing "run" entry. Unable to build Bionetwork.')
        run_dict = config['run']

        # Get network parameters
        # step time (dt) is set in the kernel and should be passed
        overwrite = run_dict['overwrite_output_dir'] if 'overwrite_output_dir' in run_dict else True
        print_time = run_dict['print_time'] if 'print_time' in run_dict else False
        dt = run_dict['dt']  # TODO: make sure dt exists
        tstop = float(config.tstop) / 1000.0
        network = cls(graph, dt=config.dt, tstop=tstop, overwrite=overwrite)

        if 'output_dir' in config['output']:
            network.output_dir = config['output']['output_dir']

        # network.spikes_file = config['output']['spikes_ascii']

        if 'block_run' in run_dict and run_dict['block_run']:
            if 'block_size' not in run_dict:
                raise Exception('"block_run" is set to True but "block_size" not found.')
            network._block_size = run_dict['block_size']

        if 'duration' in run_dict:
            network.duration = run_dict['duration']

        graph.io.log_info('Building cells.')
        graph.build_nodes()

        graph.io.log_info('Building recurrent connections')
        graph.build_recurrent_edges()

        for sim_input in inputs.from_config(config):
            node_set = graph.get_node_set(sim_input.node_set)
            if sim_input.input_type == 'spikes':
                path = sim_input.params['input_file']
                spikes = SpikeTrains.load(path=path, file_type=sim_input.module, **sim_input.params)
                graph.io.log_info('Build virtual cell stimulations for {}'.format(sim_input.name))
                graph.add_spike_trains(spikes, node_set)
            else:
                graph.io.log_info('Build virtual cell stimulations for {}'.format(sim_input.name))
                rates = firing_rates.RatesInput(sim_input.params)
                graph.add_rates(rates, node_set)

        # Create the output file
        if 'output' in config:
            out_dict = config['output']

            rates_file = out_dict.get('rates_file', None)
            if rates_file is not None:
                rates_file = rates_file if os.path.isabs(rates_file) else os.path.join(config.output_dir, rates_file)
                # create directory if required
                network.rates_file = rates_file
                parent_dir = os.path.dirname(rates_file)
                if not os.path.exists(parent_dir):
                    os.makedirs(parent_dir)

            if 'log_file' in out_dict:
                log_file = out_dict['log_file']
                network.set_logging(log_file)

        graph.io.log_info('Network created.')
        return network



# Copyright 2020. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import types
import six
import warnings
from functools import wraps


class _PyFunctions(object):
    """Structure that allows users to add functions that will overwrite certain parts of bmtk, in particular; how
    cell-objects are created, how cells are processed, how synapatic objects are created, and how synaptic weights are
    calculated. To use the user must write a custom function in their own scripts and register it using the add()
    method or decorator. ex run_bionet.py::
        from bmtk.simulator.bionet import synaptic_weight

        @synaptic_weight
        def calc_custom_weight(edge_props, src_cell, trg_cell):
            syn_weight = ... # calulate updated weight using edge and cells props
            return syn_weight

    This will register the user function "calc_custom_weight(...)" in the py_modules object. Any edge object that has
    attribute weight_function=="calc_custom_weight" will use this new user-defined function to readjust weights. ie
    in the bmtk code where syn_weight value is calculate for each edge::
        from bmtk.simulator.bionet.pyfunction_cache import py_modules

        if 'weight_function' in edge_props:
            fnc_name = edge_props['weight_function']
            assert(py_modules.has_synaptic_weight(name=fnc_name)
            weight_fnc = py_modules.synaptic_weight(name=fnc_name)
            syn_weight = weight_fnc(edge_props, src_cell, trg_cell)
        else:
            syn_weight = ...  # default way to get weight


    The following functionality can be overwritten by the user.
        cell_model: function to create a NEURON/NEST/DiPDE cell object(s). Uses SONATA properties "model_template" and
            "model_type".

        cell_processor: Used to do some post-processing of cell object after being created, ex fixing part of morphology
            or updating channel variable parameters. Uses "model_processing" property.

        synapse_model: Used to create a synapse between two (sets of) cells. Uses SONATA edges "model_template"
            property.

        synaptic_weight: Used to update/recalculate synaptic weight of an edge (synapse or junction). Uses
            "weight_function" (not a default SONATA variable).
    """
    def __init__(self):
        self.__syn_weights = {}
        self.__cell_models = {}
        self.__synapse_models = {}
        self.__cell_processors = {}

    def clear(self):
        self.__syn_weights.clear()
        self.__cell_models.clear()
        self.__synapse_models.clear()
        self.__cell_processors.clear()

    def add_synaptic_weight(self, name, func, overwrite=True):
        """stores synaptic function for given name"""
        if overwrite or name not in self.__syn_weights:
            self.__syn_weights[name] = func

    @property
    def synaptic_weights(self):
        """return list of the names of all available synaptic weight functions"""
        return self.__syn_weights.keys()

    def synaptic_weight(self, name):
        """return the synaptic weight function"""
        return self.__syn_weights[name]

    def has_synaptic_weight(self, name):
        return name in self.__syn_weights

    def __cell_model_key(self, directive, model_type):
        return (directive, model_type)

    def add_cell_model(self, directive, model_type, func, overwrite=True):
        if not model_type:
            # use * to indicate "model_type" is not specified and therefore a wildcard option.
            model_type = '*'

        key = self.__cell_model_key(directive, model_type)
        if overwrite or key not in self.__cell_models:
            self.__cell_models[key] = func

    @property
    def cell_models(self):
        return self.__cell_models.keys()

    def cell_model(self, directive, model_type='*'):
        # Check to see if function exists with corresponding "directive" and "model_type". If not see if there is a
        # function with "directive" but "model_type=*" to act as fall-through option
        if self.has_cell_model(directive=directive, model_type=model_type):
            return self.__cell_models[self.__cell_model_key(directive, model_type)]

        elif model_type != '*' and self.has_cell_model(directive=directive, model_type='*'):
            return self.__cell_models[self.__cell_model_key(directive, '*')]

        else:
            raise ValueError('Could not find cell_model() function with directive="{}" and model_type="{}"'.format(
                directive, model_type
            ))

    def has_cell_model(self, directive, model_type='*'):
        if not model_type:
            model_type = '*'
        return self.__cell_model_key(directive, model_type) in self.__cell_models

    def add_synapse_model(self, name, func, overwrite=True):
        if overwrite or name not in self.__synapse_models:
            self.__synapse_models[name] = func

    def has_synapse_model(self, name):
        return name in self.__synapse_models

    @property
    def synapse_models(self):
        return self.__synapse_models.keys()

    def synapse_model(self, name):
        return self.__synapse_models[name]

    def has_cell_processor(self, name):
        return name in self.__cell_processors

    @property
    def cell_processors(self):
        return self.__cell_processors.keys()

    def cell_processor(self, name):
        return self.__cell_processors[name]

    def add_cell_processor(self, name, func, overwrite=True):
        if overwrite or name not in self.__syn_weights:
            self.__cell_processors[name] = func

    def __repr__(self):
        rstr = '{}: {}\n'.format('cell_models', self.cell_models)
        rstr += '{}: {}\n'.format('synapse_models', self.synapse_models)
        rstr += '{}: {}\n'.format('synaptic_weights', self.synaptic_weights)
        rstr += '{}: {}'.format('cell_processors', self.cell_processors)
        return rstr


py_modules = _PyFunctions()


def synaptic_weight(*wargs, **wkwargs):
    """A decorator for registering a function as a synaptic weight function.
    To use either::

        @synaptic_weight
        def weight_function():
            ...

    or::

        @synaptic_weight(name='name_in_edge_types')
        def weight_function():
            ...

    Once the decorator has been attached and imported the functions will automatically be added to py_modules and BMTK
    will when assigning synaptic/gap junction weights for edges with matching "weight_function" attribute
    """
    if len(wargs) == 1 and callable(wargs[0]):
        # for the case without decorator arguments, grab the function object in wargs and create a decorator
        func = wargs[0]
        py_modules.add_synaptic_weight(func.__name__, func)  # add function assigned to its original name

        @wraps(func)
        def func_wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return func_wrapper
    else:
        # for the case with decorator arguments
        assert(all(k in ['name'] for k in wkwargs.keys()))
        def decorator(func):
            # store the function in py_modules but under the name given in the decorator arguments
            py_modules.add_synaptic_weight(wkwargs['name'], func)

            @wraps(func)
            def func_wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return func_wrapper
        return decorator


def cell_model(*wargs, **wkwargs):
    """A decorator for registering NEURON cell loader functions."""
    if len(wargs) == 1 and callable(wargs[0]):
        # for the case without decorator arguments, grab the function object in wargs and create a decorator
        func = wargs[0]
        py_modules.add_cell_model(
            directive=func.__name__,  # add function assigned to its original name
            model_type='*',
            func=func
        )

        @wraps(func)
        def func_wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return func_wrapper
    else:
        # look for directive/name argument inside decorator arguments, otherwise default
        if 'directive' in wkwargs:
            directive = wkwargs['directive']
        elif 'name' in wkwargs:
            directive = wkwargs['name']
        elif len(wargs) >= 1 and isinstance(wargs[0], six.string_types):
            directive = wargs[0]
        elif len(wargs) >= 1 and callable(wargs[0]):
            directive = wargs[0].__name__
        else:
            raise ValueError('Please specify "directive" name in call_model() arguments')

        model_type = wkwargs.get('model_type', '*')
        overwrite = wkwargs.get('overwrite', True)

        def decorator(func):
            # store the function in py_modules but under the name given in the decorator arguments
            py_modules.add_cell_model(directive=directive, model_type=model_type, func=func, overwrite=overwrite)

            @wraps(func)
            def func_wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return func_wrapper
        return decorator


def synapse_model(*wargs, **wkwargs):
    """A decorator for registering NEURON synapse loader functions."""
    if len(wargs) == 1 and callable(wargs[0]):
        # for the case without decorator arguments, grab the function object in wargs and create a decorator
        func = wargs[0]
        py_modules.add_synapse_model(func.__name__, func)  # add function assigned to its original name

        @wraps(func)
        def func_wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return func_wrapper
    else:
        # for the case with decorator arguments
        assert(all(k in ['name'] for k in wkwargs.keys()))

        def decorator(func):
            # store the function in py_modules but under the name given in the decorator arguments
            py_modules.add_synapse_model(wkwargs['name'], func)

            @wraps(func)
            def func_wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return func_wrapper
        return decorator


def add_weight_function(func, name=None, overwrite=True):
    assert(callable(func))
    func_name = name if name is not None else func.__name__
    py_modules.add_synaptic_weight(func_name, func, overwrite)


def add_cell_model(func, directive, model_type='*', overwrite=True):
    assert(callable(func))
    py_modules.add_cell_model(directive=directive, model_type=model_type, func=func, overwrite=overwrite)


def add_cell_processor(func, name=None, overwrite=True):
    assert(callable(func))
    func_name = name if name is not None else func.__name__
    py_modules.add_cell_processor(func_name, func, overwrite)


def add_synapse_model(func, name=None, overwrite=True):
    assert (callable(func))
    func_name = name if name is not None else func.__name__
    py_modules.add_synapse_model(func_name, func, overwrite)


def load_py_modules(cell_models=None, syn_models=None, syn_weights=None, cell_processors=None):
    # py_modules.clear()
    warnings.warn('Do not call this method directly', DeprecationWarning)
    if cell_models is not None:
        assert(isinstance(cell_models, types.ModuleType))
        for f in [cell_models.__dict__.get(f) for f in dir(cell_models)]:
            if isinstance(f, types.FunctionType):
                py_modules.add_cell_model(f.__name__, f)

    if syn_models is not None:
        assert(isinstance(syn_models, types.ModuleType))
        for f in [syn_models.__dict__.get(f) for f in dir(syn_models)]:
            if isinstance(f, types.FunctionType):
                py_modules.add_synapse_model(f.__name__, f)

    if syn_weights is not None:
        assert(isinstance(syn_weights, types.ModuleType))
        for f in [syn_weights.__dict__.get(f) for f in dir(syn_weights)]:
            if isinstance(f, types.FunctionType):
                py_modules.add_synaptic_weight(f.__name__, f)

    if cell_processors is not None:
        assert(isinstance(cell_processors, types.ModuleType))
        for f in [cell_processors.__dict__.get(f) for f in dir(cell_processors)]:
            if isinstance(f, types.FunctionType):
                py_modules.add_cell_processor(f.__name__, f)



import csv
import numpy as np
from six import string_types

from bmtk.simulator.core.simulator import Simulator
import bmtk.simulator.utils.simulation_inputs as inputs
from bmtk.simulator.filternet.config import Config
from bmtk.simulator.filternet.lgnmodel.movie import *
from bmtk.simulator.filternet import modules as mods
from bmtk.simulator.filternet.io_tools import io
from bmtk.utils.io.ioutils import bmtk_world_comm


class FilterSimulator(Simulator):
    def __init__(self, network, dt, tstop):
        super(FilterSimulator, self).__init__()
        self._network = network
        self._dt = dt
        self._tstop = tstop/1000.0
        self._io = network.io

        self.rates_csv = None
        self._movies = []
        self._eval_options = []

    @property
    def io(self):
        return self._io

    @property
    def dt(self):
        return self._dt

    def add_movie(self, movie_type, params):
        # TODO: Move this into its own factory
        movie_type = movie_type.lower() if isinstance(movie_type, string_types) else 'movie'
        if movie_type == 'movie' or not movie_type:
            if 'data_file' in params:
                m_data = None
                if 'data_file' in params:
                    m_data = np.load(params['data_file'])
                elif 'data' in params:
                    m_data = params['data']
                else:
                    raise Exception('Could not find movie "data_file" in config to use as input.')

                contrast_min, contrast_max = m_data.min(), m_data.max()
                normalize_data = params.get('normalize', False)
                if contrast_min < -1.0 or contrast_max > 1.0:
                    if normalize_data:
                        self.io.log_info('Normalizing movie data to (-1.0, 1.0).')
                        m_data = m_data*2.0/(contrast_max - contrast_min) - 1.0
                    else:
                        self.io.log_info('Movie data range is not normalized to (-1.0, 1.0).')

                init_params = FilterSimulator.find_params(['row_range', 'col_range', 'labels', 'units', 'frame_rate',
                                                           't_range'], **params)
                self._movies.append(Movie(m_data, **init_params))

        elif movie_type == 'full_field':
            raise NotImplementedError

        elif movie_type == 'full_field_flash':
            init_params = FilterSimulator.find_params(['row_size', 'col_size', 't_on', 't_off', 'max_intensity',
                                                       'frame_rate'], **params)
            init_params['row_range'] = range(init_params['row_size'])
            del init_params['row_size']
            init_params['col_range'] = range(init_params['col_size'])
            del init_params['col_size']
            init_params['t_on'] = init_params['t_on']/1000.0
            init_params['t_off'] = init_params['t_off']/1000.0
            init_params['max_intensity'] = init_params.get('max_intensity', 1)# *-1.0

            ffm = FullFieldFlashMovie(**init_params)
            mv = ffm.full(t_max=self._tstop)
            self._movies.append(mv)

        elif movie_type == 'graiting':
            init_params = FilterSimulator.find_params(['row_size', 'col_size', 'frame_rate'], **params)
            create_params = FilterSimulator.find_params(['gray_screen_dur', 'cpd', 'temporal_f', 'theta', 'contrast'],
                                                        **params)

            create_params['gray_screen_dur'] /= 1000.0
            gm = GratingMovie(**init_params)
            graiting_movie = gm.create_movie(t_min=0.0, t_max=self._tstop, **create_params)
            self._movies.append(graiting_movie)

        elif movie_type == 'looming':
            init_params = FilterSimulator.find_params(['row_size', 'col_size', 'frame_rate'], **params)
            movie_params = FilterSimulator.find_params(['t_looming', 'gray_sceen_dur'], **params)
            lm = LoomingMovie(**init_params)
            looming_movie = lm.create_movie(**movie_params)
            self._movies.append(looming_movie)

        else:
            raise Exception('Unknown movie type {}'.format(movie_type))

        if 'evaluation_options' in params:
            self._eval_options.append(params['evaluation_options'])
        else:
            self._eval_options.append({})

    def run(self):
        for mod in self._sim_mods:
            mod.initialize(self)

        io.log_info('Evaluating rates.')

        cells_on_rank = self.local_cells()
        n_cells_on_rank = len(cells_on_rank)
        ten_percent = int(np.ceil(n_cells_on_rank*0.1))
        rank_msg = '' if bmtk_world_comm.MPI_size < 2 else ' (on rank {})'.format(bmtk_world_comm.MPI_rank)

        for cell_num, cell in enumerate(cells_on_rank):
            for movie, options in zip(self._movies, self._eval_options):
                if cell_num > 0 and cell_num % ten_percent == 0:
                    io.log_debug(' Processing cell {} of {}{}.'.format(cell_num, n_cells_on_rank, rank_msg))
                ts, f_rates = cell.lgn_cell_obj.evaluate(movie, **options)

                for mod in self._sim_mods:
                    mod.save(self, cell, ts, f_rates)

        io.log_info('Done.')
        for mod in self._sim_mods:
            mod.finalize(self)

    def local_cells(self):
        return self._network.cells()

    @staticmethod
    def find_params(param_names, **kwargs):
        ret_dict = {}
        for pn in param_names:
            if pn in kwargs:
                ret_dict[pn] = kwargs[pn]

        return ret_dict

    @classmethod
    def from_config(cls, config, network):
        if not isinstance(config, Config):
            try:
                config = Config.load(config, False)
            except Exception as e:
                network.io.log_exception('Could not convert {} (type "{}") to json.'.format(config, type(config)))

        if not config.with_networks:
            network.io.log_exception('Could not find any network files. Unable to build network.')

        sim = cls(network=network, dt=config.dt, tstop=config.tstop)

        if config.jitter is not None:
            network.jitter = config.jitter

        for sim_input in inputs.from_config(config):
            if sim_input.input_type == 'movie':
                sim.add_movie(sim_input.module, sim_input.params)
            else:
                raise Exception('Unable to load input type {}'.format(sim_input.input_type))

        network.io.log_info('Building cells.')
        network.build_nodes()

        rates_csv = config.output.get('rates_csv', None)
        rates_h5 = config.output.get('rates_h5', None)
        if rates_csv or rates_h5:
            sim.add_mod(mods.RecordRates(rates_csv, rates_h5, config.output_dir))

        spikes_csv = config.output.get('spikes_csv', None) or config.output.get('spikes_file_csv', None)
        spikes_h5 = config.output.get('spikes_h5', None) or config.output.get('spikes_file', None)
        spikes_nwb = config.output.get('spikes_nwb', None) or config.output.get('spikes_file_nwb', None)
        if spikes_csv or spikes_h5 or spikes_nwb:
            sim.add_mod(mods.SpikesGenerator(spikes_csv, spikes_h5, spikes_nwb, config.output_dir))

        return sim



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
from . import utils
from .file_root import NodesRoot, EdgesRoot


class File(object):
    def __init__(self, data_files, data_type_files, mode='r', gid_table=None, require_magic=True):
        if mode != 'r':
            raise Exception('Currently only read mode is supported.')

        self._data_files = utils.listify(data_files)
        self._data_type_files = utils.listify(data_type_files)

        # Open and check HDF5 file(s)
        self._h5_file_handles = [utils.load_h5(f, mode) for f in self._data_files]
        if require_magic:
            map(utils.check_magic, self._h5_file_handles)  # Check magic attribute in h5 files

        # Check version number
        avail_versions = set(map(utils.get_version, self._h5_file_handles))
        if len(avail_versions) == 1:
            self._version = list(avail_versions)[0]
        elif len(avail_versions) > 1:
            # TODO: log as warning
            print('Warning: Passing in multiple hdf5 files of different version')
            self._version = ','.join(avail_versions)
        else:
            self._version = utils.VERSION_NA

        self._csv_file_handles = [(f, utils.load_csv(f)) for f in self._data_type_files]

        self._has_nodes = False
        self._nodes = None  # /nodes object
        self._nodes_groups = []  # list of all hdf5 /nodes group
        self._node_types_dataframes = []  # list of all csv node-types dataframe

        self._has_edges = False
        self._edges = None  # /edges object
        self._edges_groups = []  # list of all hdf5 /edges group
        self._edge_types_dataframes = []  # list of csv edge-types dataframes

        # for multiple inputs sort into edge files and node files
        self._sort_types_file()
        self._sort_h5_files()

        if not (self._has_nodes or self._has_edges):
            raise Exception('Could not find neither nodes nor edges for the given file(s).')

        if self._has_nodes:
            self._nodes = NodesRoot(nodes=self._nodes_groups, node_types=self._node_types_dataframes, gid_table=gid_table)

        if self._has_edges:
            self._edges = EdgesRoot(edges=self._edges_groups, edge_types=self._edge_types_dataframes)

    @property
    def nodes(self):
        return self._nodes

    @property
    def has_nodes(self):
        return self._has_nodes

    @property
    def edges(self):
        return self._edges

    @property
    def has_edges(self):
        return self._has_edges

    @property
    def version(self):
        return self._version

    def _sort_types_file(self):
        # TODO: node/edge type_id columnn names should not be hardcoded
        for filename, df in self._csv_file_handles:
            has_node_type_id = 'node_type_id' in df.columns
            has_edge_type_id = 'edge_type_id' in df.columns
            if has_node_type_id and has_edge_type_id:
                # TODO: users may be creating their own dataframe and thus not have a filename
                raise Exception('types file {} has both node_types_id and edge_types_id column.'.format(filename))
            elif has_node_type_id:
                self._node_types_dataframes.append(df)
            elif has_edge_type_id:
                self._edge_types_dataframes.append(df)
            else:
                # TODO: if strict this should fail immedietely
                print('Warning: Could not determine if file {} was an edge-types or node-types file. Ignoring'.format(filename))

    def _sort_h5_files(self):
        for h5 in self._h5_file_handles:
            has_nodes = '/nodes' in h5
            has_edges = '/edges' in h5
            if not (has_nodes or has_edges):
                print('File {} contains neither nodes nor edges. Ignoring'.format(h5.filename))
            else:
                if has_nodes:
                    self._nodes_groups.append(h5)
                    self._has_nodes = True
                if has_edges:
                    self._edges_groups.append(h5)
                    self._has_edges = True



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


class SimNode(object):
    def __init__(self, node_id, graph, network, params):
        self._node_id = node_id
        self._graph = graph
        self._graph_params = params
        self._node_type_id = params['node_type_id']
        self._network = network
        self._updated_params = {}

        self._model_params = {}

    @property
    def node_id(self):
        return self._node_id

    @property
    def node_type_id(self):
        return self._node_type_id

    @property
    def network(self):
        """Name of network node belongs too."""
        return self._network

    @property
    def model_params(self):
        """Parameters (json file, nml, dictionary) that describe a specific node"""
        return self._model_params

    @model_params.setter
    def model_params(self, value):
        self._model_params = value

    def __contains__(self, item):
        return item in self._updated_params or item in self._graph_params

    def __getitem__(self, item):
        if item in self._updated_params:
            return self._updated_params[item]
        else:
            return self._graph_params[item]


class PopNode(SimNode):
    def __init__(self, node_id, graph, network, params):
        self._graph = graph
        self._node_id = node_id
        self._network = network
        self._graph_params = params

        self._dynamics_params = {}
        self._updated_params = {'dynamics_params': self._dynamics_params}

        self._gids = set()

    @property
    def node_id(self):
        return self._node_id

    @property
    def pop_id(self):
        return self._node_id

    @property
    def network(self):
        return self._network

    @property
    def dynamics_params(self):
        return self._dynamics_params

    @dynamics_params.setter
    def dynamics_params(self, value):
        self._dynamics_params = value

    @property
    def is_internal(self):
        return False

    def __getitem__(self, item):
        if item in self._updated_params:
            return self._updated_params[item]
        elif item in self._graph_params:
            return self._graph_params[item]
        elif self._model_params is not None:
            return self._model_params[item]

    def add_gid(self, gid):
        self._gids.add(gid)

    def get_gids(self):
        return list(self._gids)


class InternalNode(PopNode):
    @property
    def tau_m(self):
        return self['tau_m']

    @tau_m.setter
    def tau_m(self, value):
        #return self['tau_m']
        self._dynamics_params['tau_m'] = value

    @property
    def v_max(self):
        return self._dynamics_params.get('v_max', None)

    @v_max.setter
    def v_max(self, value):
        self._dynamics_params['v_max'] = value

    @property
    def dv(self):
        return self._dynamics_params.get('dv', None)

    @dv.setter
    def dv(self, value):
        self._dynamics_params['dv'] = value

    @property
    def v_min(self):
        return self._dynamics_params.get('v_min', None)

    @v_min.setter
    def v_min(self, value):
        self._dynamics_params['v_min'] = value

    @property
    def is_internal(self):
        return True

    def __repr__(self):
        props = 'pop_id={}, tau_m={}, v_max={}, v_min={}, dv={}'.format(self.pop_id, self.tau_m, self.v_max, self.v_min,
                                                                        self.dv)
        return 'InternalPopulation({})'.format(props)


class ExternalPopulation(PopNode):
    def __init__(self, node_id, graph, network, params):
        super(ExternalPopulation, self).__init__(node_id, graph, network, params)
        self._firing_rate = -1
        if 'firing_rate' in params:
            self._firing_rate = params['firing_rate']

    @property
    def firing_rate(self):
        return self._firing_rate

    @property
    def is_firing_rate_set(self):
        return self._firing_rate >= 0

    @firing_rate.setter
    def firing_rate(self, rate):
        assert(isinstance(rate, float) and rate >= 0)
        self._firing_rate = rate

    @property
    def is_internal(self):
        return False

    def __repr__(self):
        props = 'pop_id={}, firing_rate={}'.format(self.pop_id, self.firing_rate)
        return 'ExternalPopulation({})'.format(props)




import logging

from . import env_builder


logger = logging.getLogger(__name__)


environment_builders = {
    'bionet': env_builder.BioNetEnvBuilder,
    'biophysical': env_builder.BioNetEnvBuilder,
    'bio': env_builder.BioNetEnvBuilder,
    'pointnet': env_builder.PointNetEnvBuilder,
    'point': env_builder.PointNetEnvBuilder,
    'filternet': env_builder.FilterNetEnvBuilder,
    'filter': env_builder.FilterNetEnvBuilder,
    'popnet': env_builder.PopNetEnvBuilder,
    'pop': env_builder.PopNetEnvBuilder
}


def create_environment(simulator,
                       base_dir='.',
                       network_dir=None,
                       components_dir=None,
                       node_sets_file=None,
                       output_dir=None,
                       overwrite=False,
                       run_script=True,
                       config_file=None,
                       config_name=None,
                       split_configs=False,
                       network_filter=None,
                       report_vars=[],
                       report_nodes=None,
                       spikes_inputs=None,
                       rates_inputs=None,
                       clamp_reports=[],
                       current_clamp=None,
                       file_current_clamp=None,
                       se_voltage_clamp=None,
                       tstart=0.0,
                       tstop=1000.0,
                       dt=0.001,
                       dL=20.0,
                       spikes_threshold=-15.0,
                       nsteps_block=5000,
                       v_init=-80.0,
                       celsius=34.0,
                       compile_mechanisms=False,
                       use_relative_paths=True,
                       include_examples=False
                       ):
    """

    :param simulator:
    :param base_dir:
    :param network_dir:
    :param components_dir:
    :param node_sets_file:
    :param output_dir:
    :param overwrite:
    :param run_script:
    :param config_file:
    :param config_name:
    :param split_configs:
    :param network_filter:
    :param report_vars:
    :param report_nodes:
    :param spikes_inputs:
    :param rates_inputs:
    :param clamp_reports:
    :param current_clamp:
    :param file_current_clamp:
    :param se_voltage_clamp:
    :param tstart:
    :param tstop:
    :param dt:
    :param dL:
    :param spikes_threshold:
    :param nstesp_block:
    :param v_init:
    :param celsius:
    :param compile_mechanisms:
    :param use_relative_paths:
    :param include_examples:
    """
    simulator_name = simulator.lower()
    if simulator_name not in environment_builders.keys():
        logging.error('Could not find bmtk simulator {}; available options: {}'.format(
            simulator_name, list(environment_builders.keys())
        ))

    build_cls = environment_builders[simulator_name]
    env_builder = build_cls(
        base_dir=base_dir,
        network_dir=network_dir,
        components_dir=components_dir,
        output_dir=output_dir,
        node_sets_file=node_sets_file
    )

    env_builder.build(
        include_examples=include_examples,
        use_relative_paths=use_relative_paths,
        report_vars=report_vars,
        report_nodes=report_nodes,
        clamp_reports=clamp_reports,
        current_clamp=current_clamp,
        file_current_clamp=file_current_clamp,
        se_voltage_clamp=se_voltage_clamp,
        spikes_inputs=spikes_inputs,
        rates_inputs=rates_inputs,
        tstart=tstart,
        tstop=tstop,
        dt=dt,
        dL=dL,
        spikes_threshold=spikes_threshold,
        nsteps_block=nsteps_block,
        v_init=v_init,
        celsius=celsius,
        config_file=config_file,
        overwrite_config=overwrite,
        config_name=config_name,
        split_configs=split_configs,
        network_filter=network_filter,
        run_script=run_script
    )

    if compile_mechanisms:
        env_builder.compile_mechanisms()



import matplotlib.pyplot as plt
import numpy as np
import copy
import warnings

from .compartment_report import CompartmentReport


def __get_population(report, population):
    """Helper function to figure out which population of nodes to use."""
    pops = report.populations
    if population is None:
        # If only one population exists in spikes object/file select that one
        if len(pops) > 1:
            raise Exception('Compartment Report contains more than one population of nodes. Use "population" parameter '
                            'to specify population to display.')

        else:
            return pops[0]

    elif population not in pops:
        raise Exception('Could not find node population "{}" in , only found {}'.format(population, pops))

    else:
        return population

def __get_node_groups(report, node_groups, population):
    """Helper function for parsing the 'node_groups' params"""
    if node_groups is None:
        # If none are specified by user make a 'node_group' consisting of all nodes
        selected_nodes = report.node_ids(population=population)
        return [{'node_ids': selected_nodes, 'c': 'b', 'label': 'averaged'}], selected_nodes
    else:
        # Fetch all node_ids which can be used to filter the data.
        node_groups = copy.deepcopy(node_groups)  # Make a copy since later we may be altering the dictionary
        selected_nodes = np.array(node_groups[0]['node_ids'])
        for grp in node_groups[1:]:
            if 'node_ids' not in grp:
                raise AttributeError('Could not find "node_ids" key in node_groups parameter.')
            selected_nodes = np.concatenate((selected_nodes, np.array(grp['node_ids'])))

        return node_groups, selected_nodes


def plot_traces(report, population=None, node_ids=None, sections='origin', average=False, node_groups=None, times=None,
                title=None, show_legend=None, show=True, save_as=None, plt_style=None):
    """Displays the time trace of one or more nodes from a SONATA CompartmentReport file.

    To plot a group of individual variable traces (based on their soma)::

       plot_traces('/path/to/report.h5', node_ids=[0, 10, 20, ...])

    If node_ids=None (default) then all nodes in the report will be displayed. For large networks then can become
    difficult to visualize so it's recommended you use average=True::

        plot_traces('/path/to/report.h5', average=True)

    Users also have the option of taking the averages of multiple subsets of nodes using the "node_groups" options.
    "node_groups" should be a list of dictionary, each dict with a 'node_ids': [list of ids], and optionally
    a 'label' and 'c' (color). For example support nodes [0, 70) are excitatory cells, nodes [70, 100) are inhibitory,
    and we want display excitatory and inhibitory averages and blue and red, respectivly::

        plot_traces('/path/to/report.h5',
                    node_groups=[{'node_ids': range(0, 70), 'label': 'exc', 'c': 'b'},
                                 {'node_ids': range(70, 100), 'label': 'inh', 'c': 'r'}])

    :param report: Path to SONATA report file or CompartmentReport object
    :param population: string. If the report more than one population of nodes, use this to determine which nodes to
           plot. If only one population exists and population=None then the function will find it by default.
    :param node_ids: int or list of ints. Individual node to display the variable
    :param sections: 'origin', 'all', or list of ids, Compartments/elements to display, By default will only show values
           at the soma.
    :param average: If True will take the averages of all/selected nodes. Default False
    :param node_groups: None or list of dicts. Used to group sets of nodes by labels and color. Each grouping should
        be a dictionary with a 'node_ids' key with a list of the ids. You can also add 'label' and 'c' keys for
        label and color. If None all nodes will be labeled and colored the same.
    :param times: (float, float), start and stop times of simulation
    :param title: str, adds a title to the plot
    :param show_legend: Set True or False to determine if legend should be displayed on the plot. The default (None)
           function itself will guess if legend should be shown.
    :param show: bool to display or not display plot. default True.
    :param save_as: None or str: file-name/path to save the plot as a png/jpeg/etc. If None or empty string will not
        save plot.
    :return: matplotlib figure.Figure object
    """
    if plt_style is not None:
        plt.style.use(plt_style)

    if node_groups is not None and node_ids is not None:
        warnings.warn('plot_traces is called with both "node_ids" and "node_groups" parameters.', UserWarning)

    elif node_ids is not None:
        if average and (not np.isscalar(node_ids)) and len(node_ids) > 1:
            node_groups = {'node_ids': node_ids, 'label': average}
            return plot_traces_averaged(report=report, population=population, sections=sections,
                                        node_groups=node_groups, times=times, title=title, show_legend=show_legend,
                                        show=show, save_as=save_as)

        return plot_traces_individual(report=report, population=population, node_ids=node_ids, sections=sections,
                                      times=times, title=title, show_legend=show_legend, show=show, save_as=save_as)

    elif node_groups is not None:
        return plot_traces_averaged(report=report, population=population, sections=sections, node_groups=node_groups,
                                    times=times, title=title, show_legend=show_legend, show=show, save_as=save_as)

    elif average is not None:
        return plot_traces_averaged(report=report, population=population, sections=sections, times=times, title=title,
                                    show_legend=show_legend, show=show, save_as=save_as)

    else:
        return plot_traces_individual(report=report, population=population, sections=sections, times=times, title=title,
                                      show_legend=show_legend, show=show, save_as=save_as)


def plot_traces_individual(report, population=None, node_ids=None, sections='origin', times=None, title=None,
                           show_legend=None, show=True, save_as=None):
    """Used the plot time traces of individual nodes from a SONATA compartment report file.

    Recommended use plot_traces instead, which will call this function if necessarcy.

    :param report: Path to SONATA report file or CompartmentReport object.
    :param population: string. If the report more than one population of nodes, use this to determine which nodes to
           plot. If only one population exists and population=None then the function will find it by default.
    :param node_ids: int or list of ints. Individual node to display the variable.
    :param sections: 'origin', 'all', or list of ids, Compartments/elements to display, By default will only show values
           at the soma.
    :param times: (float, float), start and stop times of simulation.
    :param title: str, adds a title to the plot.
    :param show_legend: Set True or False to determine if legend should be displayed on the plot. The default (None)
           function itself will guess if legend should be shown.
    :param show: bool to display or not display plot. default True.
    :param save_as: None or str: file-name/path to save the plot as a png/jpeg/etc. If None or empty string will not
        save plot.
    :return: matplotlib figure.Figure object
    """

    cr = CompartmentReport.load(report)
    pop = __get_population(report=cr, population=population)  # get node populations name

    if node_ids is None:
        node_ids = cr.node_ids(population=pop)
    elif np.isscalar(node_ids):
        node_ids = [node_ids]

    trace_times = cr.time_trace(population=pop)
    if times is not None:
        times_indx = np.argwhere((trace_times >= times[0]) & (trace_times <= times[1]))
        trace_times = trace_times[times_indx]
    else:
        # Get start and stop time from the traces data
        times = (trace_times[0], trace_times[-1])

    fig, axes = plt.subplots()
    for node_id in node_ids:
        trace_data = cr.data(node_id=node_id, sections=sections, population=pop, time_window=times)
        axes.plot(trace_times, trace_data, label=node_id)

    axes.set_xlim(times[0], times[1])
    axes.set_xlabel('time ({})'.format('ms'))

    if (show_legend is None and len(node_ids) <= 10) or show_legend:
        axes.legend()

    if title:
        axes.set_title(title)

    if save_as:
        plt.savefig(save_as)

    if show:
        plt.show()

    return fig


def plot_traces_averaged(report, population=None, sections='origin', node_groups=None, times=None, title=None,
                         show_background=True, show_legend=None, show=True, save_as=None):
    """Used to plot averages across multiple nodes in a SONATA Compartment Report file.

    Recommended that you use "plot_traces" function.

    To plot multiple averages use the "node_groups" options. "node_groups" should be a list of dictionary, each dict
    with a 'node_ids': [list of ids], and optionally a 'label' and 'c' (color). For example support nodes [0, 70) are
    excitatory cells, nodes [70, 100) are inhibitory, and we want display excitatory and inhibitory averages and blue
    and red, respectively::

        plot_traces('/path/to/report.h5',
                    node_groups=[{'node_ids': range(0, 70), 'label': 'exc', 'c': 'b'},
                                 {'node_ids': range(70, 100), 'label': 'inh', 'c': 'r'}])

    :param report: Path to SONATA report file or CompartmentReport object.
    :param population: string. If the report more than one population of nodes, use this to determine which nodes to
           plot. If only one population exists and population=None then the function will find it by default.
    :param sections: 'origin', 'all', or list of ids, Compartments/elements to display, By default will only show values
           at the soma.
    :param node_groups: None or list of dicts. Used to group sets of nodes by labels and color. Each grouping should
        be a dictionary with a 'node_ids' key with a list of the ids. You can also add 'label' and 'c' keys for
        label and color. If None all nodes will be labeled and colored the same.
    :param times: (float, float), start and stop times of simulation.
    :param title: str, adds a title to the plot.
    :param show_background: shows all the individual traces greyed in the background.
    :param show_legend: Set True or False to determine if legend should be displayed on the plot. The default (None)
           function itself will guess if legend should be shown.
    :param show: bool to display or not display plot. default True.
    :param save_as: None or str: file-name/path to save the plot as a png/jpeg/etc. If None or empty string will not
        save plot.
    :return: matplotlib figure.Figure object
    """
    cr = CompartmentReport.load(report)
    pop = __get_population(report=cr, population=population)  # get node populations name
    node_groups, selected_ids = __get_node_groups(report=cr, node_groups=node_groups, population=pop)

    nodes_array = cr.node_ids(population=pop)
    trace_times = cr.time_trace(population=pop)
    traces_data = cr.data(sections=sections, population=pop)

    if times is not None:
        min_ts, max_ts = times[0], times[1]
        times_indx = np.argwhere((trace_times >= min_ts) & (trace_times <= max_ts))
        times_indx = times_indx.flatten()
        traces_data = traces_data[times_indx, :]
        trace_times = trace_times[times_indx]
    else:
        min_ts, max_ts = trace_times[0], trace_times[-1]

    fig, axes = plt.subplots()
    has_labels = False

    if show_background:
        nodes_indx = np.argwhere(np.isin(nodes_array, selected_ids))
        nodes_indx = nodes_indx.flatten()
        background_data = traces_data[:, nodes_indx]
        axes.plot(trace_times, background_data, c='lightgray')

    for node_grp in node_groups:
        grp_ids = node_grp.pop('node_ids')
        has_labels = has_labels or 'label' in node_grp

        nodes_indx = np.argwhere(np.isin(nodes_array, grp_ids))
        nodes_indx = nodes_indx.flatten()
        if len(nodes_indx) == 0:
            continue

        grp_data = traces_data[:, nodes_indx]
        grp_mean = grp_data.mean(axis=1)
        axes.plot(trace_times, grp_mean, **node_grp)

    axes.set_xlim(min_ts, max_ts)
    axes.set_xlabel('time ({})'.format('ms'))

    y_label = cr.variable(population=pop)
    if cr.units(population=pop) is not None:
        y_label += '({})'.format(cr.units(population=pop))
    axes.set_ylabel(cr.variable(population=pop))

    if title:
        axes.set_title(title)

    if (show_legend is None or show_legend) and has_labels:
        axes.legend()

    if save_as:
        plt.savefig(save_as)

    if show:
        plt.show()

    return fig



import os
import json

from bmtk.simulator.core.simulation_config import SimulationConfig
from bmtk.simulator.filternet.io_tools import io


class Config(SimulationConfig):
    @property
    def io(self):
        if self._io is None:
            self._io = io
        return self._io

    @property
    def jitter(self):
        conds = self.conditions
        has_lj = 'jitter_lower' in conds
        has_uj = 'jitter_upper' in conds

        if has_lj and has_uj:
            return (conds['jitter_lower'], conds['jitter_upper'])
        elif has_lj ^ has_uj:
            raise Exception('Please define both jitter_upper and jitter_lower parameters')
        else:
            return None



import os
import h5py
import pandas as pd
import numpy as np
import json
import shutil
import logging
import glob

from bmtk.utils.sonata.utils import add_hdf5_magic, add_hdf5_version


logger = logging.getLogger(__name__)


class ProgressFile(object):
    """Class for keeping track of the progress of the sorting of the hdf5 edges files. Will write progress to disk, so
    that if sorting fails we can still continue from where it left off."""

    def __init__(self, cache_dir, n_edges, n_chunks, sort_key, root_name=None):
        self._conf_parser = {}
        self.root_name = root_name
        self.cache_dir = cache_dir

        self.progress_file = os.path.join(self.cache_dir, 'progress.json')
        if os.path.exists(self.progress_file):
            self._conf_parser = json.load(open(self.progress_file, 'r'))
            self.initialized = self._conf_parser[self.root_name]['initialized']
            self.n_edges = self._conf_parser[self.root_name]['n_edges']
            self.n_chunks = self._conf_parser[self.root_name]['n_chunks']
            self.iteration = self._conf_parser[self.root_name]['iteration']
            self.sort_key = self._conf_parser[self.root_name]['sort_key']
            self.chunk_files = self._conf_parser[self.root_name]['chunk_files']
            self.chunk_indices = self._conf_parser[self.root_name]['chunk_indices']
            self.write_index = self._conf_parser[self.root_name]['write_index']
            self.completed = self._conf_parser[self.root_name].get('completed', False)

            # Make sure original edges file or the chunking size hasn't changed since the progress was last saved, if
            # so it can cause unforseen issues.
            if self.completed:
                raise ValueError('Already marked as complete.')

            if self.n_edges != n_edges:
                raise ValueError('n_edges={} has changed to {}. Unable to continue sorting from last saved point.'.format(self.n_edges, n_edges))

            if self.n_chunks != n_chunks:
                raise ValueError('n_chunks={} has changed to {}. Unable to continue sorting from last saved point.'.format(self.n_chunks, n_chunks))

            if self.sort_key != sort_key:
                raise ValueError('sort_key={} has changed to {}. Unable to continue from last saved point.'.format(
                    self.sort_key, sort_key
                ))

        else:
            self._conf_parser[self.root_name] = {}
            self.initialized = False
            self.n_edges = n_edges
            self.n_chunks = n_chunks
            self.iteration = 0
            self.sort_key = sort_key
            self.chunk_files = []
            self.chunk_indices = []
            self.write_index = 0
            self.completed = False
            self.update()

    def _can_serialize(self):
        json.dumps(self._conf_parser)
        return True

    def update(self, initialized=None, n_edges=None, n_chunks=None, iteration=None, chunk_files=None,
               chunk_indices=None, write_index=None, completed=None):
        if initialized is not None:
            self.initialized = initialized
        if n_edges is not None:
            self.n_edges = n_edges
        if n_chunks is not None:
            self.n_chunks = n_chunks
        if iteration is not None:
            self.iteration = iteration
        if chunk_files is not None:
            self.chunk_files = chunk_files
        if chunk_indices is not None:
            self.chunk_indices = chunk_indices
        if write_index is not None:
            self.write_index = int(write_index)
        if completed is not None:
            self.completed = completed

        self._conf_parser[self.root_name]['initialized'] = self.initialized
        self._conf_parser[self.root_name]['n_edges'] = self.n_edges
        self._conf_parser[self.root_name]['n_chunks'] = self.n_chunks
        self._conf_parser[self.root_name]['iteration'] = self.iteration
        self._conf_parser[self.root_name]['sort_key'] = self.sort_key
        self._conf_parser[self.root_name]['chunk_files'] = self.chunk_files
        self._conf_parser[self.root_name]['chunk_indices'] = self.chunk_indices
        self._conf_parser[self.root_name]['write_index'] = self.write_index
        self._conf_parser[self.root_name]['completed'] = self.completed

        if self._can_serialize():
            json.dump(self._conf_parser, open(self.progress_file, 'w'), indent=4)
        else:
            raise ValueError('Unable to serialize to json')


def _get_chunk_fn(cache_dir, itr_num, chunk_num):
    return os.path.join(cache_dir, 'itr{}_chunk{}.h5'.format(itr_num, chunk_num))


def _copy_attributes(in_grp, out_grp):
    """Recursively copy hdf5 Group/Dataset attributes from in_grp to out_grp

    :param in_grp: hdf5 Group object whose attributes will be copied from.
    :param out_grp: hdf5 Group object that will have it's attributes updated/copied to.
    """
    if in_grp.attrs:
        out_grp.attrs.update(in_grp.attrs)

    for in_name, in_h5_obj in in_grp.items():
        if in_name not in out_grp:
            # make sure matching subgroup/dataset exists in the output group
            continue

        elif isinstance(in_h5_obj, h5py.Dataset):
            out_grp[in_name].attrs.update(in_h5_obj.attrs)

        elif isinstance(in_h5_obj, h5py.Group):
            _copy_attributes(in_h5_obj, out_grp[in_name])


def _create_output_h5(input_file, output_file, edges_root, n_edges):
    mode = 'r+' if os.path.exists(output_file) else 'w'
    out_h5 = h5py.File(output_file, mode=mode)
    root_grp = out_h5.create_group(edges_root) if edges_root not in out_h5 else out_h5[edges_root]

    if 'source_node_id' not in root_grp:
        root_grp.create_dataset('source_node_id', (n_edges, ), dtype=np.uint32)

    if 'target_node_id' not in root_grp:
        root_grp.create_dataset('target_node_id', (n_edges, ), dtype=np.uint32)

    if 'edge_type_id' not in root_grp:
        root_grp.create_dataset('edge_type_id', (n_edges, ), dtype=np.uint32)

    if 'edge_group_id' not in root_grp:
        root_grp.create_dataset('edge_group_id', (n_edges, ), dtype=np.uint32)

    if 'edge_group_index' not in root_grp:
        root_grp.create_dataset('edge_group_index', (n_edges, ), dtype=np.uint32)

    # with h5py.File(input_file, 'r') as in_h5:
    #     for h5obj in in_h5[edges_root].values():
    #         if isinstance(h5obj, h5py.Group) and h5obj.name not in root_grp:
    #             root_grp.copy(h5obj, h5obj.name)

    _copy_attributes(h5py.File(input_file, 'r'), out_h5)

    return root_grp


def _order_model_groups(input_edges_path, output_edges_path, edges_population, chunk_size):
    """

    :param input_edges_path:
    :param output_edges_path:
    :param edges_population:
    :param chunk_size:
    :return:
    """
    # The output hdf5 should already have edges_group_id, edges_group_index columns which we can use to find order. This
    # hdf5 file handle should only be for reading

    # Get the model parameters from the input (original) edges file.
    in_edges_h5 = h5py.File(input_edges_path, 'r')
    in_edges_grp = in_edges_h5[edges_population]

    with h5py.File(output_edges_path, 'r+') as out_h5:
        out_root_grp = out_h5[edges_population]

        # Get a list of all model group columns that need to be resorted, create the dataset in the output hdf5 edges
        # file if it doesn't exist
        model_grp_tracker = {}
        for h5name, h5obj in in_edges_grp.items():
            if isinstance(h5obj, h5py.Group) and h5obj.name not in ['indices', 'indicies']:
                model_grp_tracker[h5name] = {'c_indx': 0, 'cols': []}
                for col_name, col_ds in h5obj.items():
                    model_grp_tracker[h5name]['cols'].append(col_name)
                    if col_ds.name not in out_h5:
                        out_h5.create_dataset(col_ds.name, shape=col_ds.shape, dtype=col_ds.dtype)

        # use edge_group_id and edge_group_index to sort the model group columns
        edge_group_ids_ds = out_root_grp['edge_group_id']
        edge_group_indices_ds = out_root_grp['edge_group_index']
        n_edges = len(edge_group_indices_ds)
        if 'edge_group_index_sorted' not in out_root_grp:
            # we will need to also need to keep track of the reordering and replace edge_group_index
            out_root_grp.create_dataset(
                'edge_group_index_sorted',
                shape=edge_group_indices_ds.shape,
                dtype=edge_group_indices_ds.dtype
            )

        chunk_idx_beg = 0
        while chunk_idx_beg < n_edges:
            # assuming we can't load and sort the entire edge_group_id/edge_group_index tables into memory, we need to
            # reorder into handable chunks

            chunk_idx_end = np.min((chunk_idx_beg + chunk_size, n_edges)).astype(np.uint)
            if chunk_idx_beg == chunk_idx_end:
                continue

            chunk_grp_ids = edge_group_ids_ds[chunk_idx_beg:chunk_idx_end][()]
            chunk_grp_idxs = edge_group_indices_ds[chunk_idx_beg:chunk_idx_end][()]
            updated_order = np.zeros(int(chunk_idx_end-chunk_idx_beg), dtype=np.uint32)

            # Go through all the groups and all their columns and reorder by group_id only
            for group_id in np.unique(chunk_grp_ids):
                # the order they appear in group_index is the order the columns will be resorted
                group_id_mask = np.argwhere(chunk_grp_ids == group_id).flatten()
                new_index_order = chunk_grp_idxs[group_id_mask]

                # If new_index_order is not sorted h5py will complain when we try to fetch those values in the index.
                # Use nio_incremental to fetch from the datasets and nio_argwhere to put back into the right order
                nio_incremental = np.sort(new_index_order)
                nio_argswhere = np.argsort(new_index_order)

                # for each column take values from in_edges and write them to out_edges in the correct order
                group_indx_beg = model_grp_tracker[str(group_id)]['c_indx']
                group_indx_end = group_indx_beg + len(new_index_order)
                for col in model_grp_tracker[str(group_id)]['cols']:
                    col_vals_tmp = in_edges_grp[str(group_id)][col][nio_incremental]
                    col_vals = col_vals_tmp[nio_argswhere]
                    out_root_grp[str(group_id)][col][group_indx_beg:group_indx_end] = col_vals

                model_grp_tracker[str(group_id)]['c_indx'] = group_indx_end
                updated_order[group_id_mask] = np.arange(group_indx_beg, group_indx_beg + len(new_index_order))

            out_root_grp['edge_group_index_sorted'][chunk_idx_beg:chunk_idx_end] = updated_order
            chunk_idx_beg = chunk_idx_end

        del out_root_grp['edge_group_index']
        out_root_grp['edge_group_index'] = out_root_grp['edge_group_index_sorted']
        del out_root_grp['edge_group_index_sorted']

        _copy_attributes(h5py.File(input_edges_path, 'r'), out_h5)


def _clean(progress):
    for c_fn in glob.glob(os.path.join(progress.cache_dir, 'itr*_chunk*.h5')):
        try:
            os.remove(c_fn)
        except Exception as e:
            pass

    try:
        os.remove(progress.progress_file)
    except Exception as e:
        pass

    try:
        shutil.rmtree(progress.cache_dir)
    except Exception as e:
        pass


def _initialize_chunks(edges_file_path, progress):
    """ Splits the original hdf5 edges file into n_chunks individual files"""
    edges_h5 = h5py.File(edges_file_path, 'r')
    edges_root_grp = edges_h5[progress.root_name]

    sort_key = progress.sort_key
    n_edges = progress.n_edges
    n_chunks = progress.n_chunks
    cache_dir = progress.cache_dir

    logger.debug('Splitting {} into {} sorted chunks'.format(edges_file_path, n_chunks))
    chunk_size = np.ceil(n_edges / n_chunks).astype(np.uint)

    # Split the edges table into N chunks, each chunk will be sorted (by the specified column) and the sorted table
    # will be saved it's own tmp hdf5 file.
    # Note: we don't have to keep track of subgroups since the edge_group_id w/
    # edge_group_index column will still hold the valid line references
    chunk_idx_beg = 0
    chunk_idx_end = np.min((chunk_size, n_edges)).astype(np.uint)
    iteration = 0
    chunk_num = 0
    chunk_files = []
    chunk_indices = []
    while chunk_idx_beg < n_edges:
        logger.debug('  Chunk {} of {} [{:,} - {:,})'.format(chunk_num+1, n_chunks, chunk_idx_beg, chunk_idx_end))
        # TODO: should be more efficencent if we just sort the specified column and keep track of the sort_order
        #   index for later.
        chunk_data_df = pd.DataFrame({
            'source_node_id': edges_root_grp['source_node_id'][chunk_idx_beg:chunk_idx_end],
            'target_node_id': edges_root_grp['target_node_id'][chunk_idx_beg:chunk_idx_end],
            'edge_type_id': edges_root_grp['edge_type_id'][chunk_idx_beg:chunk_idx_end],
            'edge_group_id': edges_root_grp['edge_group_id'][chunk_idx_beg:chunk_idx_end],
            'edge_group_index': edges_root_grp['edge_group_index'][chunk_idx_beg:chunk_idx_end]
        }).sort_values(sort_key)

        chunk_file = os.path.join(cache_dir, 'itr{}_chunk{}.h5'.format(iteration, chunk_num))
        chunk_h5 = h5py.File(chunk_file, 'w')
        chunk_h5.create_dataset('source_node_id', data=chunk_data_df['source_node_id'])
        chunk_h5.create_dataset('target_node_id', data=chunk_data_df['target_node_id'])
        chunk_h5.create_dataset('edge_type_id', data=chunk_data_df['edge_type_id'])
        chunk_h5.create_dataset('edge_group_id', data=chunk_data_df['edge_group_id'])
        chunk_h5.create_dataset('edge_group_index', data=chunk_data_df['edge_group_index'])

        # store the min/max sort-column value found in current chunk
        chunk_h5.attrs['min_id'] = chunk_data_df[sort_key].min()
        max_id = chunk_data_df[sort_key].max()
        chunk_h5.attrs['max_id'] = max_id

        # For each unique id that is being sorted on, keep a list of how many times the id shows up in the current
        # tmp chunk file
        id_counts = chunk_data_df[sort_key].value_counts().sort_index()
        id_counts = id_counts.reindex(pd.RangeIndex(max_id+1), fill_value=0)
        chunk_h5.create_dataset('id_counts', data=id_counts, dtype=np.uint32)

        chunk_files.append(chunk_file)
        chunk_indices.append(0) # chunk index is 0 b/c none of the data has been merged into the final file

        chunk_idx_beg = chunk_idx_end
        chunk_idx_end += np.min((chunk_size, n_edges)).astype(np.uint)
        chunk_num += 1

    progress.update(initialized=True, chunk_files=chunk_files, chunk_indices=chunk_indices)


def _sort_chunks(progress):
    """Takes the M (usually N or N-1) sorted chunk files, splits and merges them into N new chunk files"""
    logger.debug('Resorting Chunks')

    cache_dir = progress.cache_dir
    itr_num = progress.iteration
    n_chunks = progress.n_chunks

    # during the previous merge() process some of the rows of last iteration's sorted chunk file may have been merged
    # into the final result. Use the chunk_indices variable to keep track of what row to use for creating the next
    # sorted chunk files.
    chunk_files = progress.chunk_files
    chunk_h5s = [h5py.File(cf, 'r') for cf in chunk_files]
    chunk_start_indices = progress.chunk_indices
    chunk_sizes = [h5['source_node_id'].shape[0] for h5 in chunk_h5s]
    chunk_indices = {fn: np.linspace(c_start, c_size, num=n_chunks+1, endpoint=True, dtype=np.uint)
                     for fn, c_start, c_size in zip(chunk_files, chunk_start_indices, chunk_sizes)}

    for l in chunk_indices.values():
        assert(len(l) == n_chunks+1)

    new_chunk_files = []
    new_chunk_indices = np.zeros(n_chunks, dtype=np.uint).tolist()

    # loop through all the sorted chunk files created during the last iteration and divide them up into further chunks.
    # Take and merge together the corresponding chunks of data into a new table, sort and save for the next round of
    # merging,
    for chunk_num in range(n_chunks):
        logger.debug('  Chunk {} of {}'.format(chunk_num+1, n_chunks))
        combined_df = None
        for i, chunk_fn in enumerate(chunk_files):
            chunk_h5 = chunk_h5s[i]
            chunk_idx_beg = chunk_indices[chunk_fn][chunk_num]
            chunk_idx_end = chunk_indices[chunk_fn][chunk_num+1]
            if chunk_idx_end == chunk_idx_beg:
                continue

            data_df = pd.DataFrame({
                'source_node_id': chunk_h5['source_node_id'][chunk_idx_beg:chunk_idx_end],
                'target_node_id': chunk_h5['target_node_id'][chunk_idx_beg:chunk_idx_end],
                'edge_type_id': chunk_h5['edge_type_id'][chunk_idx_beg:chunk_idx_end],
                'edge_group_id': chunk_h5['edge_group_id'][chunk_idx_beg:chunk_idx_end],
                'edge_group_index': chunk_h5['edge_group_index'][chunk_idx_beg:chunk_idx_end]
            })
            combined_df = data_df if combined_df is None else pd.concat((combined_df, data_df))

        if combined_df is None:
            continue

        combined_df = combined_df.sort_values(progress.sort_key)

        new_chunk_file = os.path.join(cache_dir, 'itr{}_chunk{}.h5'.format(itr_num+1, chunk_num))
        new_chunk_h5 = h5py.File(new_chunk_file, 'w')
        new_chunk_h5.create_dataset('source_node_id', data=combined_df['source_node_id'])
        new_chunk_h5.create_dataset('target_node_id', data=combined_df['target_node_id'])
        new_chunk_h5.create_dataset('edge_type_id', data=combined_df['edge_type_id'])
        new_chunk_h5.create_dataset('edge_group_id', data=combined_df['edge_group_id'])
        new_chunk_h5.create_dataset('edge_group_index', data=combined_df['edge_group_index'])
        new_chunk_h5.attrs['min_id'] = combined_df[progress.sort_key].min()

        max_id = combined_df[progress.sort_key].max()
        new_chunk_h5.attrs['max_id'] = max_id

        # For each unique id that is being sorted on, keep a list of how many times the id shows up in the current
        # tmp chunk file
        id_counts = combined_df['edge_type_id'].value_counts().sort_index()
        id_counts = id_counts.reindex(pd.RangeIndex(max_id+1), fill_value=0)
        new_chunk_h5.create_dataset('id_counts', data=id_counts, dtype=np.uint32)

        new_chunk_files.append(new_chunk_file)

    progress.update(iteration=itr_num+1, chunk_files=new_chunk_files, chunk_indices=new_chunk_indices)

    logger.debug('Deleteing old chunks')
    for c_fn in chunk_files:
        try:
            os.remove(c_fn)
        except Exception as e:
            logger.warning('Could not delete file {}'.format(c_fn))
            print(e)


def _merge(progress, edges_grp):
    """Will merge data from  one or more sorted chunk files into the output file"""

    # Create a table of min and max ids in each chunk file and sort. Using this table we can find overlaps, and when
    # the max/max ids of one file doesn't overlap with another then it's safe to push that table to the output h5
    max_size = np.ceil(progress.n_edges / progress.n_chunks).astype(np.uint)
    chunk_files = progress.chunk_files.copy()
    chunk_indices = progress.chunk_indices.copy()
    chunk_nums = []
    chunk_h5s = []
    chunk_ids_min = []
    chunk_ids_max = []
    min_id = np.inf
    max_id = 0
    for chunk_num, chunk_file in enumerate(chunk_files):
        h5 = h5py.File(chunk_file, 'r+')
        chunk_nums.append(chunk_num)
        chunk_h5s.append(h5)
        chunk_ids_min.append(h5.attrs['min_id'])
        chunk_ids_max.append(h5.attrs['max_id'])
        max_id = int(np.max((max_id, h5.attrs['max_id'])))
        min_id = int(np.min((min_id, h5.attrs['min_id'])))

    logger.debug('Current status:')
    logger.debug(pd.DataFrame({
        'chunk file': [os.path.basename(c) for c in chunk_files],
        'min_id': chunk_ids_min,
        'max_id': chunk_ids_max,
    }))

    offsets = np.zeros(max_id+1, dtype=np.uint32)
    for h5 in chunk_h5s:
        id_counts = np.array(h5['id_counts'])
        offsets[:len(id_counts)] += id_counts

    offsets = np.cumsum(offsets)
    id_end = np.max((np.searchsorted(offsets, max_size, side='right'), min_id))

    collected_df = None
    for chunk_num, chunk_h5 in zip(chunk_nums, chunk_h5s):
        read_beg_idx = chunk_indices[chunk_num]
        sort_col = np.array(chunk_h5[progress.sort_key][read_beg_idx:])
        id_indices = np.argwhere(sort_col <= id_end).flatten()
        if len(id_indices) == 0:
            continue

        read_end_idx = id_indices[-1] + 1
        chunk_size = chunk_h5[progress.sort_key].shape[0]

        data_df = pd.DataFrame({
            'source_node_id': chunk_h5['source_node_id'][read_beg_idx:read_end_idx],
            'target_node_id': chunk_h5['target_node_id'][read_beg_idx:read_end_idx],
            'edge_type_id': chunk_h5['edge_type_id'][read_beg_idx:read_end_idx],
            'edge_group_id': chunk_h5['edge_group_id'][read_beg_idx:read_end_idx],
            'edge_group_index': chunk_h5['edge_group_index'][read_beg_idx:read_end_idx]
        })
        collected_df = data_df if collected_df is None else pd.concat((collected_df, data_df))

        if read_end_idx == chunk_size:
            chunk_indices[chunk_num] = -1
            chunk_files[chunk_num] = None

        else:
            chunk_indices[chunk_num] = int(read_end_idx)

    if collected_df is None or len(collected_df) == 0:
        logger.debug('No ids collected.')
        return

    collected_df = collected_df.sort_values(progress.sort_key)
    n_collected = len(collected_df)

    write_beg_idx = progress.write_index
    write_end_idx = write_beg_idx + n_collected

    logger.debug('Writing ids [{:,} - {:,}] to output indices [{:,} - {:,})'.format(
        collected_df[progress.sort_key].iloc[0],
        collected_df[progress.sort_key].iloc[-1],
        write_beg_idx,
        write_end_idx
    ))

    edges_grp['source_node_id'][write_beg_idx:write_end_idx] = collected_df['source_node_id'].values
    edges_grp['target_node_id'][write_beg_idx:write_end_idx] = collected_df['target_node_id'].values
    edges_grp['edge_type_id'][write_beg_idx:write_end_idx] = collected_df['edge_type_id'].values
    edges_grp['edge_group_id'][write_beg_idx:write_end_idx] = collected_df['edge_group_id'].values
    edges_grp['edge_group_index'][write_beg_idx:write_end_idx] = collected_df['edge_group_index'].values

    chunk_files = [c for c in chunk_files if c is not None]
    chunk_indices = [c for c in chunk_indices if c != -1]

    progress.update(chunk_files=chunk_files, chunk_indices=chunk_indices, write_index=write_end_idx,
                    completed=len(chunk_files) == 0)


def external_merge_sort(input_edges_path, output_edges_path, edges_population, sort_by, sort_model_properties=True,
                        n_chunks=12, max_itrs=10, cache_dir='.sort_cache', **kwargs):
    """Does an external merge sort on an input edges hdf5 file, saves value in new file. Usefull for large network
    files where we are not able to load into memory.

    Will split the original hdf5 into <n_chunks> chunks on the disk in cache_dir, Will sort each individual chunk
    of data in memory, then perform a merge on all the chunks. For speed considers may try to do the chunking and
    merging in multiple iterations.

    Itermediate sorting results are saved in cache_dir (eg .sort_cache) and if the sorting fails or doesn't finish in
    max_itrs, running this function again will continue where it last left off.

    :param input_edges_path: path to original edges file
    :param output_edges_path: path name of new file that will be created
    :param edges_population:
    :param sort_by: 'edge_type_id', 'source_node_id', etc.
    :param sort_model_properties: resort the model group so edges_group_id+edge_group_index is in order
    :param n_chunks: Number of chunks, eg the fraction of the edges.h5 file that will be loaded into memory at a given
        time. (default: 12)
    :param cache_dir: A temporary directory where itermeidate results will be stored. (default: './cache_dir/)
    :param max_itrs: The maximum number of iterations to run the merge sort.
    """
    with h5py.File(input_edges_path, 'r') as input_h5:
        n_edges = input_h5[edges_population]['source_node_id'].shape[0]

    cache_dir = os.path.join(os.path.dirname(output_edges_path), cache_dir,
                             os.path.splitext(os.path.basename(output_edges_path))[0])
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    n_chunks = np.min((n_chunks, n_edges))
    progress = ProgressFile(cache_dir=cache_dir, n_edges=n_edges, n_chunks=int(n_chunks), root_name=edges_population,
                            sort_key=sort_by)

    output_root_grp = _create_output_h5(input_file=input_edges_path, output_file=output_edges_path,
                                        edges_root=edges_population, n_edges=n_edges)

    # Split the edges files into N chunks.
    if not progress.initialized or not os.path.exists(cache_dir):
        _initialize_chunks(edges_file_path=input_edges_path, progress=progress)
    else:
        logger.debug('Edges files has already been initialized. Continuing from last saved point')

    # Try to merge chunk data to output, if there's still data left split and resort into another N chunks. Repeat.
    _merge(progress, edges_grp=output_root_grp)
    for i in range(max_itrs):
        if progress.completed:
            break
        _sort_chunks(progress)
        _merge(progress, edges_grp=output_root_grp)

    output_root_grp.file.flush()
    output_root_grp.file.close()

    if progress.completed and sort_model_properties:
        # copy over model group, and reorder so edge_group_ids/edge_group_index is ordered
        logger.debug('Sorting model group columns')
        chunk_size = np.max((np.ceil(n_edges / n_chunks), 2)).astype(np.uint)
        _order_model_groups(
            input_edges_path=input_edges_path,
            output_edges_path=output_edges_path,
            edges_population=progress.root_name,
            chunk_size=chunk_size
        )
    else:
        # Copy over model group columns without sorting
        out_h5 = h5py.File(output_edges_path, 'r+')
        root_grp = out_h5[progress.root_name]

        with h5py.File(input_edges_path, 'r') as in_h5:
            for h5obj in in_h5[progress.root_name].values():
                if isinstance(h5obj, h5py.Group) and h5obj.name not in root_grp and h5obj.name not in ['indices', 'indicies']:
                    root_grp.copy(h5obj, h5obj.name)

    logger.debug('Cleaning up cache directory')
    _clean(progress)

    logger.debug('Done sorting.')


if __name__ == '__main__':
    external_merge_sort(
        input_edges_path='network/right_local_edges.h5',
        output_edges_path='network/right_local_edges_by_edgetype_merge_sort_v1.h5',
        sort_by='edge_type_id',
        edges_population='/edges/right_local'
    )



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import h5py
import numpy as np
import os
import datetime


def load_spikes(file_name,trial_name=None):
    """Loads spikes from multiple file formats

    :param file_name: name of spike file
    :param trian_name: name of a trial within a spike file

    :return spikes: dict with spikes
    """

    filename, file_extension = os.path.splitext(file_name)

    if file_extension == ".nwb":
        spikes = load_spikes_from_nwb(file_name,trial_name)

    if file_extension == ".txt":
        spikes = load_spikes_from_txt(file_name)

    if file_extension == ".h5":
        spikes = load_spikes_from_h5(file_name)


    return spikes

    
def load_spikes_from_txt(file_name):
    """Load spikes from the txt file
    
    :param file_name: name of spike file
    :param trian_name: name of a trial within a spike file

    :return spikes: dict with spikes
    """
    t = os.path.getmtime(file_name)
    spk_ts, spk_gids = np.loadtxt(file_name, dtype='float32,int', unpack=True)
    spikes = {
        "time": spk_ts,
        "gid": spk_gids
    }
    print('loaded spikes from txt')
    return spikes


def load_spikes_from_h5(file_name):
    """Load spikes from the hdf5 file
    
    :param file_name: name of spike file
    :param trian_name: name of a trial within a spike file

    :return spikes: dict with spikes
    """
    t = os.path.getmtime(file_name)
    print("{} modified on: {}".format(file_name, datetime.datetime.fromtimestamp(t)))

    spikes ={}
    with h5py.File(file_name,'r') as h5:
        spikes["time"] = h5["time"][...]
        spikes["gid"] = h5["gid"][...]

    return spikes


def load_spikes_from_nwb(file_name,trial_name):
    """Load spikes from the nwb file
    
    :param file_name: name of spike file
    :param trian_name: name of a trial within a spike file

    :return spikes: dict with spikes
    """
    f5 = h5py.File(file_name, 'r')

    spike_trains_handle = f5['processing/%s/spike_train' % trial_name]  # nwb.SpikeTrain.get_processing(f5,'trial_0')
    spike_times = []
    spike_gids = []
    for gid in spike_trains_handle.keys():
        times_gid = spike_trains_handle['%d/data' %int(gid)][:]
        spike_times.extend(times_gid)
        spike_gids.extend([int(gid)]*len(times_gid))
        
    spikes = {}
    spikes["time"] = np.array(spike_times) 
    spikes["gid"] = np.array(spike_gids)

    return spikes



from .kernel import Kernel2D, Kernel3D
from .cursor import LNUnitCursor, MultiLNUnitCursor, MultiLNUnitMultiMovieCursor, SeparableLNUnitCursor, \
    SeparableMultiLNUnitCursor

    
class LNUnit(object):
    def __init__(self, linear_filter, transfer_function, amplitude=1.):
        self.linear_filter = linear_filter
        self.transfer_function = transfer_function
        self.amplitude = amplitude

    def evaluate(self, movie, **kwargs):
        return self.get_cursor(movie, separable=kwargs.pop('separable', False)).evaluate(**kwargs)
 
    def get_spatiotemporal_kernel(self, *args, **kwargs):
        return self.linear_filter.get_spatiotemporal_kernel(*args, **kwargs)
    
    def get_cursor(self, movie, threshold=0, separable=False):
        if separable:
            return SeparableLNUnitCursor(self, movie)
        else:
            return LNUnitCursor(self, movie, threshold=threshold)
    
    def show_temporal_filter(self, *args, **kwargs):
        self.linear_filter.show_temporal_filter(*args, **kwargs)
        
    def show_spatial_filter(self, *args, **kwargs):
        self.linear_filter.show_spatial_filter(*args, **kwargs)
    
    def to_dict(self):
        return {
            'class': (__name__, self.__class__.__name__),
            'linear_filter': self.linear_filter.to_dict(),
            'transfer_function': self.transfer_function.to_dict()
        }


class MultiLNUnit(object):
    def __init__(self, lnunit_list, transfer_function):
        self.lnunit_list = lnunit_list
        self.transfer_function = transfer_function
        
    def get_spatiotemporal_kernel(self, *args, **kwargs):
        k = Kernel3D([], [], [], [], [], [], [])
        for unit in self.lnunit_list:
            k = k+unit.get_spatiotemporal_kernel(*args, **kwargs)
            
        return k
        
    def show_temporal_filter(self, *args, **kwargs):
        import matplotlib.pyplot as plt
        ax = kwargs.pop('ax', None)
        show = kwargs.pop('show', None)
        save_file_name = kwargs.pop('save_file_name', None) 

        if ax is None:
            _, ax = plt.subplots(1, 1)
        
        kwargs.update({'ax': ax, 'show': False, 'save_file_name': None})
        for unit in self.lnunit_list:
            if unit.linear_filter.amplitude < 0:
                color = 'b'
            else:
                color = 'r'
            unit.linear_filter.show_temporal_filter(color=color, **kwargs)

        if save_file_name is not None:
            plt.savefig(save_file_name, transparent=True)
         
        if show:
            plt.show()
         
        return ax
    
    def show_spatial_filter(self, *args, **kwargs):
        
        ax = kwargs.pop('ax', None)
        show = kwargs.pop('show', True)
        save_file_name = kwargs.pop('save_file_name', None) 
        colorbar = kwargs.pop('colorbar', True) 
        
        k = Kernel2D(args[0], args[1], [], [], [])
        for lnunit in self.lnunit_list:
            k = k + lnunit.linear_filter.spatial_filter.get_kernel(*args, **kwargs)
        k.imshow(ax=ax, show=show, save_file_name=save_file_name, colorbar=colorbar)
        
    def get_cursor(self, *args, **kwargs):
        
        threshold = kwargs.get('threshold', 0.)
        separable = kwargs.get('separable', False)
        
        if len(args) == 1:
            movie = args[0]
            if separable:
                return SeparableMultiLNUnitCursor(self, movie)
            else:
                return MultiLNUnitCursor(self, movie, threshold=threshold)
        elif len(args) > 1:
            movie_list = args
            if separable:
                raise NotImplementedError
            else:
                return MultiLNUnitMultiMovieCursor(self, movie_list, threshold=threshold)
        else:
            assert ValueError
    
    def evaluate(self, movie, **kwargs):
        seperable = kwargs.pop('separable', False)
        return self.get_cursor(movie, separable=seperable).evaluate(**kwargs)



from setuptools import setup, find_packages
import io
import bmtk


package_name = 'bmtk'
package_version = bmtk.__version__


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


with open('README.md', 'r') as fhandle:
    long_description = fhandle.read()


setup(
    name=package_name,
    version=package_version,
    description='Brain Modeling Toolkit',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/AllenInstitute/bmtk',
    author='Kael Dai',
    author_email='kaeld@alleninstitute.org',
    keywords=['neuroscience', 'scientific', 'modeling', 'simulation'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],

    # setup_requires=['pytest-runner', 'flake8'],
    tests_require=['pytest'],
    install_requires=[
        'jsonschema',
        'pandas',
        'numpy',
        'six',
        'h5py',
        'matplotlib',
        'enum; python_version <= "2.7"',
        'scipy',
        'scikit-image',  # Only required for filternet, consider making optional
        'sympy'  # For FilterNet
    ],
    extras_require={
        'bionet': ['NEURON'],
        'mintnet': ['tensorflow'],
        'pointnet': ['NEST'],
        'popnet': ['DiPDE']
    },
    packages=find_packages(exclude=['bmtk.tests', 'bmtk.tests.*']),
    # package_data={'': ['*.md', '*.txt', '*.cfg', '**/*.json', '**/*.hoc']},
    include_package_data=True,
    platforms='any'

)



import logging
from optparse import OptionParser

from .compile_mechanisms import compile_mechanisms, copy_modfiles


logger = logging.getLogger(__name__)


if __name__ == '__main__':
    # logging.basicConfig(level=logging.INFO, format='%(module)s [%(levelname)s] %(message)s')
    logging.basicConfig(level=logging.INFO, format='compile_mechanisms [%(levelname)s] %(message)s')

    parser = OptionParser(usage="Usage: python %prog [--copy-modfiles] mechanisms-dir")
    parser.add_option('-c', '--copy-modfiles', dest='copy_modfiles', action='store_true', default=False,
                      help='Also copy over Allen Cell Type modfiles/ directory into mechanism-dir.')

    options, args = parser.parse_args()

    mechanisms_dir = args[0]
    if options.copy_modfiles:
        copy_modfiles(mechanisms_dir=mechanisms_dir)

    compile_mechanisms(mechanisms_dir=mechanisms_dir)



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np
import math


def positions_columinar(N=1, center=[0.0, 50.0, 0.0], height=100.0, min_radius=0.0, max_radius=1.0,
                        distribution='uniform'):
    """Returns a set of random x,y,z coordinates within a given sphere. Height is given as the y (index 1) coordates

    :param N: Number of points to return
    :param center: center of sphere
    :param height: maximum length of sphere (y coord)
    :param min_radius: minimum horizontal radius on x-z plane
    :param max_radius: maximum horizontal radius on x-z plane
    :param distribution: Not yet implementd
    :return: A (N, 3) matrix
    """
    phi = 2.0 * math.pi * np.random.random([N])
    r = np.sqrt((min_radius**2 - max_radius**2) * np.random.random([N]) + max_radius**2)
    x = center[0] + r * np.cos(phi)
    z = center[2] + r * np.sin(phi)
    y = center[1] + height * (np.random.random([N]) - 0.5)

    return np.column_stack((x, y, z))


def positions_cuboid(N=1, center=[0.0, 0.0, 0.0], height=100.0, xside_length=100.0, yside_length=100.0, min_dist=20.0):
    """This function distributes the cells in a 3D cuboid (x,y,z sides may have different lengths). The method used
    assures cells cannot be placed too close to one another (must be > min_dist apart)
    WARNING: If cell density is high and there is more than 1 population of cells, there is a high chance cells will be
    placed on top of one another. You can use positions_list() to avoid this...

    Written by Ben Latimer at University of Missouri (latimerb@missouri.edu)

    :return: A (N, 3) matrix
    """
    
    # Create the possible x,y,z coordinates
    x_grid = np.arange(center[0], xside_length+min_dist, min_dist)
    y_grid = np.arange(center[1], yside_length+min_dist, min_dist)
    z_grid = np.arange(center[2], height+min_dist, min_dist)
    xx, yy, zz = np.meshgrid(x_grid, y_grid, z_grid)
    positions = np.vstack([xx.ravel(), yy.ravel(), zz.ravel()]).T

    # Pick N indices for the positions matrix randomly, without replacement (so coordinates are unique)
    inds = np.random.choice(np.arange(0, np.size(positions, 0)), N, replace=False)

    # Assign positions
    x = positions[inds][:, 0]
    y = positions[inds][:, 1]
    z = positions[inds][:, 2]

    return np.column_stack((x, y, z))


def positions_list(positions=np.array([(0, 0, 0), (0, 0, 1)])):
    """This function is designed to be used with an externally supplied array of x,y,z coordinates. It is useful to
    avoid cell overlap in high density situations or to make some unique geometry. After you create each population,
    delete those positions from the "master" list of coordinates. This will assure that no cells are placed on top of
    one another.

    Written by Ben Latimer at University of Missouri (latimerb@missouri.edu)

    :param positions:
    :return:
    """
    
    x = positions[:, 0]
    y = positions[:, 1]
    z = positions[:, 2]

    return np.column_stack((x, y, z))



def xiter_random(N=1, min_x=0.0, max_x=1.0):
    return np.random.uniform(low=min_x, high=max_x, size=(N,))
